''' Tests for pyleoclim.core.ui.MultipleSeries

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''
import numpy as np
import pandas as pd

from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

import pytest
from urllib.request import urlopen
import json

import pyleoclim as pyleo
from pyleoclim.utils.tsmodel import (
    ar1_sim,
    colored_noise,
)
# from pyleoclim.utils.decomposition import mcpca

# a collection of useful functions

def gen_ts(model, nt, alpha, t=None):
    'wrapper for gen_ts in pyleoclim'
    t, v = pyleo.utils.gen_ts(model=model, nt=nt, alpha=alpha, t=t)
    ts = pyleo.Series(t, v)
    return ts

def gen_normal(loc=0, scale=1, nt=100):
    ''' Generate random data with a Gaussian distribution
    '''
    t = np.arange(nt)
    v = np.random.normal(loc=loc, scale=scale, size=nt)
    return t, v

def gen_colored_noise(alpha=1, nt=100, f0=None, m=None, seed=None):
    ''' Generate colored noise
    '''
    t = np.arange(nt)
    v = colored_noise(alpha=alpha, t=t, f0=f0, m=m, seed=seed)
    return t, v
    
def load_data():
    #Loads stott MD982176 record
    try:
        d = pyleo.Lipd(usr_path='http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004')
    except:
        d = pyleo.Lipd('./example_data/MD982176.Stott.2004.lpd')
    return d

# Tests below
class TestUIMultipleSeriesDetrend():
    @pytest.mark.parametrize('detrend_method',['linear','constant','savitzky-golay','emd'])
    def test_detrend_t1(self, detrend_method):
        alpha=1
        t, v = gen_colored_noise(nt=550, alpha=alpha)
        #Trends
        slope = 1e-5
        slope1= 2e-5
        intercept = -1
        nonlinear_trend = slope*t**2 + intercept
        nonlinear_trend1 = slope1*t**2 + intercept
        v_trend = v + nonlinear_trend
        v_trend1 = v + nonlinear_trend1

        #create series object
        ts = pyleo.Series(time=t,value=v_trend)
        ts1 = pyleo.Series(time=t,value=v_trend1)

        # Create a multiple series object
        ts_all= pyleo.MultipleSeries([ts,ts1])
        ts_detrend=ts_all.detrend(method=detrend_method)     
        detrend_0 = ts_detrend.series_list[0]
        detrend_1 = ts_detrend.series_list[1]

        assert len(detrend_0.value)==len(detrend_0.time)
        assert len(detrend_1.value)==len(detrend_1.time)

class TestMultipleSeriesPlot:
    '''Test for MultipleSeries.plot()

    MultipleSeries.plot outputs a matplotlib figure and axis object with two datasets,
    so we will compare the time axis of the axis object to the time arrays we generate,
    and the value axis with the value arrays we generate'''

    def test_plot(self):

        #Generate time and value arrays
        t_0, v_0 = gen_normal()
        t_1, v_1 = gen_normal()

        #Create series objects
        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        #Create a list of series objects
        serieslist = [ts_0, ts_1]

        #Turn this list into a multiple series object
        ts_M = pyleo.MultipleSeries(serieslist)

        fig, ax = ts_M.plot()

        lines_0 = ax.lines[0]
        lines_1 = ax.lines[1]

        x_plot_0 = lines_0.get_xdata()
        y_plot_0 = lines_0.get_ydata()

        x_plot_1 = lines_1.get_xdata()
        y_plot_1 = lines_1.get_ydata()

        assert_array_equal(t_0, x_plot_0)
        assert_array_equal(t_1, x_plot_1)
        assert_array_equal(v_0, y_plot_0)
        assert_array_equal(v_1, y_plot_1)

class TestMultipleSeriesStripes:
    '''Test for MultipleSeries.stripes()

    '''

    def test_stripes(self):

        #Generate time and value arrays
        t_0, v_0 = gen_normal()
        t_1, v_1 = gen_normal()

        #Create series objects
        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        #Create a list of series objects
        serieslist = [ts_0, ts_1]

        #Turn this list into a multiple series object
        ts_M = pyleo.MultipleSeries(serieslist)

        fig, ax = ts_M.stripes(LIM=4, x_offset=2, label_color = 'red')


class TestMultipleSeriesStandardize:
    '''Test for MultipleSeries.standardize()

    Standardize normalizes the multiple series object, so we'll simply test maximum and minimum values,
    only now we are running the test on series in a MultipleSeries object'''

    def test_standardize(self):
        t_0, v_0 = gen_colored_noise()
        t_1, v_1 = gen_colored_noise()

        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        serieslist = [ts_0, ts_1]

        ts_M = pyleo.MultipleSeries(serieslist)

        ts_M_std = ts_M.standardize()

        x_axis_0 = ts_M_std.series_list[0].__dict__['time']
        x_axis_1 = ts_M_std.series_list[1].__dict__['time']

        y_axis_0 = ts_M_std.series_list[0].__dict__['value']
        y_axis_1 = ts_M_std.series_list[1].__dict__['value']

        assert_array_equal(x_axis_0, t_0)
        assert_array_equal(x_axis_1, t_1)

        assert max(v_0) > max(y_axis_0)
        assert max(v_1) > max(y_axis_1)

class TestMultipleSeriesBin:
    '''Test for MultipleSeries.bin()

    Testing if the bin function will place the series on the same time axis
    '''

    def test_bin(self):
        t_0, v_0 = gen_colored_noise()
        t_1, v_1 = gen_colored_noise()

        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        serieslist = [ts_0, ts_1]

        ts_M = pyleo.MultipleSeries(serieslist)

        ts_M_bin = ts_M.bin()

        x_axis_0 = ts_M_bin.series_list[0].__dict__['time']
        x_axis_1 = ts_M_bin.series_list[1].__dict__['time']

        assert_array_equal(x_axis_0, x_axis_1)

class TestMultipleSeriesInterp:
    '''Test for MultipleSeries.interp()

    Testing if the interp function will place the series on the same time axis
    '''

    def test_interp(self):
        t_0, v_0 = gen_colored_noise()
        t_1, v_1 = gen_colored_noise()

        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        serieslist = [ts_0, ts_1]

        ts_M = pyleo.MultipleSeries(serieslist)

        ts_M_interp = ts_M.interp()

        x_axis_0 = ts_M_interp.series_list[0].__dict__['time']
        x_axis_1 = ts_M_interp.series_list[1].__dict__['time']

        assert_array_equal(x_axis_0, x_axis_1)
    
class TestMultipleSeriesGkernel:
    '''Test for MultipleSeries.gkernel()

    Testing the gkernel function will place the series on the same time axis
    '''

    def test_gkernel(self):
        t_0, v_0 = gen_colored_noise()
        t_1, v_1 = gen_colored_noise()

        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        serieslist = [ts_0, ts_1]

        ts_M = pyleo.MultipleSeries(serieslist)

        ts_M_gkernel = ts_M.gkernel()

        x_axis_0 = ts_M_gkernel.series_list[0].__dict__['time']
        x_axis_1 = ts_M_gkernel.series_list[1].__dict__['time']

        assert_array_equal(x_axis_0, x_axis_1)    

class TestMultipleSeriesPca:
    '''Tests for MultipleSeries.pca()

    Testing the PCA function 
    '''

    def test_pca_t0(self):
        '''
        Test with synthetic data, no missing values, screeplot()

        Returns
        -------
        None.

        '''
        p = 10; n = 100
        signal = gen_ts(model='colored_noise', nt=n, alpha=1.0).standardize()
        X = signal.value[:,None] + np.random.randn(n,p)
        t = np.arange(n)
    
        mslist = []
        for i in range(p):
            mslist.append(pyleo.Series(time = t, value = X[:,i]))
        ms = pyleo.MultipleSeries(mslist)

        res = ms.pca()
        
        # check that all variance was recovered
        assert abs(res.pctvar.sum() - 100)<0.1 
        
    
    def test_pca_t1(self):
        '''
        Test with synthetic data, with missing values

        '''
        p = 10; n = 100
        signal = gen_ts(model='colored_noise', nt=n, alpha=1.0).standardize()
        X = signal.value[:,None] + np.random.randn(n,p)
        t = np.arange(n)
        
        # poke some holes at random in the array
        Xflat = X.flatten()
        Xflat[np.random.randint(n*p, size=p-1)]=np.nan  # note: at most ncomp missing vals
        X = np.reshape(Xflat, (n,p))
    
        #X[-1,0] = np.nan
    
        mslist = []
        for i in range(p):
            mslist.append(pyleo.Series(time = t, value = X[:,i],clean_ts=False))
        ms = pyleo.MultipleSeries(mslist)

        res = ms.pca(ncomp=4, gls=True)
                
        fig, ax = res.screeplot()
        pyleo.closefig(fig)
        
    def test_pca_t2(self):
        '''
        Test with real data, same time axis
    
        ''' 
        d=load_data()
        tslist = d.to_LipdSeriesList()
        tslist = tslist[2:]
        ms = pyleo.MultipleSeries(tslist)
        msl = ms.common_time()  # put on common time
    
        res = msl.pca()
        
        fig, ax = res.screeplot()
        pyleo.closefig(fig)
        fig, ax = res.modeplot()
        pyleo.closefig(fig)
        
    def test_pca_t3(self):
        '''
        Test with synthetic data, no missing values, kwargs

        Returns
        -------
        None.

        '''
        p = 10; n = 100
        signal = gen_ts(model='colored_noise', nt=n, alpha=1.0)
        X = signal.value[:,None] + np.random.randn(n,p)
        t = np.arange(n)
    
        mslist = []
        for i in range(p):
            mslist.append(pyleo.Series(time = t, value = X[:,i]))
        ms = pyleo.MultipleSeries(mslist)

        res = ms.pca(method='eig',standardize=True,demean=False,normalize=True)
        # check that all variance was recovered
        assert abs(res.pctvar.sum() - 100)<0.001 
        
 
class TestMultipleSeriesIncrements:
    '''Test for MultipleSeries.increments()
    
    '''
    @pytest.mark.parametrize('step_style', ['min', 'max', 'mean', 'median'])
    def test_increments(self, step_style):
        p = 2; n = 100
        signal = gen_ts(model='colored_noise', nt=n, alpha=1.0).standardize()
        X = signal.value[:,None] + np.random.randn(n,p)
        t = np.arange(n)
    
        mslist = []
        for i in range(p):
            mslist.append(pyleo.Series(time = t, value = X[:,i]))
        ms = pyleo.MultipleSeries(mslist)
        
        gp = ms.increments(step_style=step_style)
        
        assert (gp[0,:] == np.array((t.min(), t.max(), 1.))).all()
               
# class TestMultipleSeriesMcPca:
#     '''Test for MultipleSeries.mcpca()

#     Testing the MC-PCA function 
#     '''    
#     def test_mcpca_t0(self):
#         url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
#         data = pyleo.Lipd(usr_path = url)
#         tslist = data.to_LipdSeriesList()
#         tslist = tslist[2:] # drop the first two series which only concerns age and depth
#         ms = pyleo.MultipleSeries(tslist)
    
#         # TO DO !!!!
    
#         # msc = ms.common_time()
    
#         # res = msc.pca(nMC=20)   

    
class TestMultipleSeriesCommonTime:
    '''Test for MultipleSeries.common_time()
    '''
    @pytest.mark.parametrize('method', ['bin', 'interp', 'gkernel'])
    def test_common_time_t0(self, method):
        t_0, v_0 = gen_colored_noise()
        t_1, v_1 = gen_colored_noise()

        ts_0 = pyleo.Series(time = t_0, value = v_0)
        ts_1 = pyleo.Series(time = t_1, value = v_1)

        serieslist = [ts_0, ts_1]

        ts_M = pyleo.MultipleSeries(serieslist)

        ts_M_ct = ts_M.common_time(method=method)

        x_axis_0 = ts_M_ct.series_list[0].time
        x_axis_1 = ts_M_ct.series_list[1].time

        assert_array_equal(x_axis_0, x_axis_1)
        
    def test_common_time_t1(self):
        time = np.arange(1900, 2020, step=1/12)
        ndel = 200
        seriesList = []
        n = 100
        for j in range(4):
            v = gen_ts(model='colored_noise', nt=n, alpha=1, t=time)
            deleted_idx = np.random.choice(range(np.size(time)), ndel, replace=False)
            tu =  np.delete(time.copy(), deleted_idx)
            vu =  np.delete(v.value, deleted_idx)
            ts = pyleo.Series(time=tu, value=vu,  value_name='Series_'+str(j+1))
            seriesList.append(ts)
    
        ms = pyleo.MultipleSeries(seriesList)

        ms1 = ms.common_time(method='interp', start=1910, stop=2010, step=1/12)
        
        assert (np.diff(ms1.series_list[0].time)[0] - 1/12) < 1e-3
        
class TestMultipleSeriesStackPlot():
    ''' Test for MultipleSeries.Stackplot
    '''
    
    @pytest.mark.parametrize('labels', [None, 'auto', ['sst','d18Osw']])
    def test_StackPlot_t0(self, labels):
    
        d=load_data()
        sst = d.to_LipdSeries(number=5)
        d18Osw = d.to_LipdSeries(number=3)
        ms = pyleo.MultipleSeries([sst,d18Osw])
        fig, ax = ms.stackplot(labels=labels)
        pyleo.closefig(fig)
    
    @pytest.mark.parametrize('plot_kwargs', [{'marker':'o'},[{'marker':'o'},{'marker':'^'}]])
    def test_StackPlot_t1(self, plot_kwargs):
    
        d=load_data()
        sst = d.to_LipdSeries(number=5)
        d18Osw = d.to_LipdSeries(number=3)
        ms = pyleo.MultipleSeries([sst,d18Osw])

        fig, ax = ms.stackplot(plot_kwargs=plot_kwargs)
        pyleo.closefig(fig)
        
class TestMultipleSeriesSpectral():
    ''' Test for MultipleSeries.spectral
    '''
    @pytest.mark.parametrize('spec_method',['wwz','cwt'])
    def test_spectral_t0(self,spec_method):
        '''Test the spectral function with pre-generated scalogram objects
        '''
        
        d = load_data()
        sst = d.to_LipdSeries(number=5)
        d18Osw = d.to_LipdSeries(number=3)
        if spec_method == 'cwt':
            sst = sst.interp()
            d18Osw = d18Osw.interp()
        ms = pyleo.MultipleSeries([sst,d18Osw])
        scals = ms.wavelet(method=spec_method)
        psds = ms.spectral(method=spec_method,scalogram_list=scals)
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Eigendecomposition methods:
Singular Spectrum Analysis (SSA). 
soon: Monte-Carlo Principal Component Analysis, Multi-Channel SSA
"""

__all__ = [
    'ssa'
]

import numpy as np
#from sklearn.decomposition import PCA
#from statsmodels.multivariate.pca import PCA
from .tsutils import standardize
from .tsmodel import ar1_sim
from scipy.linalg import eigh, toeplitz
#from nitime import algorithms as alg
#import copy

#------
# Main functions
#------

# def mcpca(ys, nMC=200, **pca_kwargs):
#     '''Monte-Carlo Principal Component Analysis

#     Carries out Principal Component Analysis  (most unfortunately named EOF analysis in the meteorology
#     and climate literature) on a data matrix ys.

#     The significance of eigenvalues is gauged against those of AR(1) surrogates fit to the data.

#     TODO: enable for ensembles and generally debug

#     Parameters
#     ----------
#     ys : 2D numpy array (nt, nrec)
#         nt   = number of time samples (assumed identical for all records)
#         nrec = number of records (aka variables, channels, etc)

#     nMC : int
#         the number of Monte-Carlo simulations

#     pca_kwargs : tuple
#         keyword arguments for the PCA method

#     Returns
#     -------
#     res : dict containing:

#         - eigvals : eigenvalues (nrec,)

#         - eigvals95 : eigenvalues of the AR(1) ensemble (nrec, nMC)

#         - pcs : PC series of all components (nt, rec)

#         - eofs : EOFs of all components (nrec, nrec)

#     References:
#     ----------
#     Deininger, M., McDermott, F., Mudelsee, M. et al. (2017): Coherency of late Holocene
#     European speleothem δ18O records linked to North Atlantic Ocean circulation.
#     Climate Dynamics, 49, 595–618. https://doi.org/10.1007/s00382-016-3360-8

#     Written by Jun Hu (Rice University).
#     Adapted for pyleoclim by Julien Emile-Geay (USC)
#     '''
#     nt, nrec = ys.shape

#     ncomp = min(nt,nrec)

#     pc_mc = np.zeros((nt,nrec)) # principal components
#     eof_mc = np.zeros((nrec,nrec))  #eof (spatial loadings)
#     #eigenvalue matrices
#     eigvals = np.zeros((nrec))
#     eig_ar1 = np.zeros((nrec,nMC))

#     # apply PCA algorithm to the data matrix
#     pca_res = PCA(ys,ncomp=ncomp, **pca_kwargs)
#     eigvals = pca_res.eigenvals

#     # generate surrogate matrix
#     y_ar1 = np.full((nt,nrec,nMC), 0, dtype=np.double)

#     for i in range(nrec):
#         yi = copy.deepcopy(ys[:,i])
#         # generate nMC AR(1) surrogates
#         y_ar1[:,i,:] = ar1_sim(yi, nMC)
#         # assign PC and EOF matrices
#         if pc.loadings[:,i][0]>0:
#             eof_mc[:,i]  = pc.loadings[:,i]
#             pc_mc[:,i]   = pc.factors[:,i]
#         else:   # flip sign (arbitrary)
#             eof_mc[:,i]  = -pc.loadings[:,i]
#             pc_mc[:,i]   = -pc.factors[:,i]
#         # estimate effective sample size
#         #PC1 =
#         neff[i] = tsutils.eff_sample_size(PC1)

#     # loop over Monte Carlo iterations
#     for m in range(nMC):
#         pc_ar1 = PCA(y_ar1[:,:,m],ncomp=nrec,**pca_kwargs)
#         eig_ar1[:,m] = pc_ar1.eigenvals

#     eig95 = np.percentile(eig_ar1, 95, axis=1)


#     # assign result
#     #res = {'eigvals': eigvals, 'eigvals95': eig95, 'pcs': pc_mc, 'eofs': eof_mc}

#     # compute effective sample size
#     PC1  = out.factors[:,0]
#     neff = tsutils.eff_sample_size(PC1)

#     # compute percent variance
#     pctvar = out.eigenvals**2/np.sum(out.eigenvals**2)*100

#     # assign result to SpatiamDecomp class
#     # Note: need to grab coordinates from Series or LiPDSeries
#     res = SpatialDecomp(name='PCA', time = self.series_list[0].time, neff= neff,
#                         pcs = out.scores, pctvar = pctvar,  locs = None,
#                         eigvals = out.eigenvals, eigvecs = out.eigenvecs)

#     return res


# def mssa(ys, M=None, nMC=0, f=0.3):
#     '''Multi-channel singular spectrum analysis analysis

#     Multivariate generalization of SSA [2], using the original algorithm of [1].
#     Each variable is called a channel, hence the name.

#     Parameters
#     ----------

#     ys : array
#           multiple time series (dimension: length of time series x total number of time series)

#     M : int
#        window size (embedding dimension, default: 10% of the length of the series)

#     nMC : int
#        Number of iteration in the Monte-Carlo process [default=0, no Monte Carlo process]

#     f : float
#        fraction (0<f<=1) of good data points for identifying significant PCs [f = 0.3]

#     Returns
#     -------
#     res : dict
#         Containing:

#         - eigvals : array of eigenvalue spectrum

#         - eigvals05 : The 5% percentile of eigenvalues

#         - eigvals95 : The 95% percentile of eigenvalues

#         - PC : matrix of principal components (2D array)

#         - RC : matrix of RCs (nrec,N,nrec*M) (2D array)

#     References
#     ----------
#     [1]_ Vautard, R., and M. Ghil (1989), Singular spectrum analysis in nonlinear
#     dynamics, with applications to paleoclimatic time series, Physica D, 35,
#     395–424.

#     [2]_ Jiang, N., J. D. Neelin, and M. Ghil (1995), Quasi-quadrennial and
#     quasi-biennial variability in the equatorial Pacific, Clim. Dyn., 12, 101-112.

#     See Also
#     --------

#     pyleoclim.utils.decomposition.ssa : Singular Spectrum Analysis (single channel)

#     '''
#     N = len(ys[:, 0])
#     nrec = len(ys[0, :])
#     if M == None:
#         M=int(N/10)
#     Y = np.zeros((N - M + 1, nrec * M))
#     for irec in np.arange(nrec):
#         for m in np.arange(0, M):
#             Y[:, m + irec * M] = ys[m:N - M + 1 + m, irec]

#     C = np.dot(np.nan_to_num(np.transpose(Y)), np.nan_to_num(Y)) / (N - M + 1)
#     D, eigvecs = eigh(C)

#     sort_tmp = np.sort(D)
#     eigvals = sort_tmp[::-1]
#     sortarg = np.argsort(-D)

#     eigvecs = eigvecs[:, sortarg]

#     # test the signifiance using Monte-Carlo
#     Ym = np.zeros((N - M + 1, nrec * M))
#     noise = np.zeros((nrec, N, nMC))
#     for irec in np.arange(nrec):
#         noise[irec, 0, :] = ys[0, irec]
#     eigvals_R = np.zeros((nrec * M, nMC))
#     # estimate coefficents of ar1 processes, and then generate ar1 time series (noise)
#     # TODO : update to use ar1_sim(), as in ssa()
#     for irec in np.arange(nrec):
#         Xs = ys[:, irec]
#         coefs_est, var_est = alg.AR_est_YW(Xs[~np.isnan(Xs)], 1)
#         sigma_est = np.sqrt(var_est)

#         for jt in range(1, N):
#             noise[irec, jt, :] = coefs_est * noise[irec, jt - 1, :] + sigma_est * np.random.randn(1, nMC)

#     for m in range(nMC):
#         for irec in np.arange(nrec):
#             noise[irec, :, m] = (noise[irec, :, m] - np.mean(noise[irec, :, m])) / (
#                 np.std(noise[irec, :, m], ddof=1))
#             for im in np.arange(0, M):
#                 Ym[:, im + irec * M] = noise[irec, im:N - M + 1 + im, m]
#         Cn = np.dot(np.nan_to_num(np.transpose(Ym)), np.nan_to_num(Ym)) / (N - M + 1)
#         # eigvals_R[:,m] = np.diag(np.dot(np.dot(eigvecs,Cn),np.transpose(eigvecs)))
#         eigvals_R[:, m] = np.diag(np.dot(np.dot(np.transpose(eigvecs), Cn), eigvecs))

#     eigvals95 = np.percentile(eigvals_R, 95, axis=1)
#     eigvals05 = np.percentile(eigvals_R, 5, axis=1)


#     # determine principal component time series
#     PC = np.zeros((N - M + 1, nrec * M))
#     PC[:, :] = np.nan
#     for k in np.arange(nrec * M):
#         for i in np.arange(0, N - M + 1):
#             #   modify for nan
#             prod = Y[i, :] * eigvecs[:, k]
#             ngood = sum(~np.isnan(prod))
#             #   must have at least m*f good points
#             if ngood >= M * f:
#                 PC[i, k] = sum(prod[~np.isnan(prod)])  # the columns of this matrix are Ak(t), k=1 to M (T-PCs)

#     # compute reconstructed timeseries
#     Np = N - M + 1

#     RC = np.zeros((nrec, N, nrec * M))

#     for k in np.arange(nrec):
#         for im in np.arange(M):
#             x2 = np.dot(np.expand_dims(PC[:, im], axis=1), np.expand_dims(eigvecs[0 + k * M:M + k * M, im], axis=0))
#             x2 = np.flipud(x2)

#             for n in np.arange(N):
#                 RC[k, n, im] = np.diagonal(x2, offset=-(Np - 1 - n)).mean()
#     res = {'eigvals': eigvals, 'eigvecs': eigvecs, 'q05': eigvals05, 'q95': eigvals95, 'PC': PC, 'RC': RC}

#     return res

def ssa(y, M=None, nMC=0, f=0.5, trunc=None, var_thresh = 80):
    '''Singular spectrum analysis

    Nonparametric eigendecomposition of timeseries into orthogonal oscillations.
    This implementation  uses the method of [1], with applications presented in [2].
    Optionally (nMC>0), the significance of eigenvalues is assessed by Monte-Carlo simulations of an AR(1) model fit to X, using [3].
    The method expects regular spacing, but is tolerant to missing values, up to a fraction 0<f<1 (see [4]).

    Parameters
    ----------

    y : array of length N
          time series (evenly-spaced, possibly with up to f*N NaNs)

    M : int
       window size (default: 10% of N)

    nMC : int
        Number of iterations in the Monte-Carlo simulation (default nMC=0, bypasses Monte Carlo SSA)
        Currently only supported for evenly-spaced, gap-free data.

    f : float
        maximum allowable fraction of missing values. (Default is 0.5)

    trunc : str
        if present, truncates the expansion to a level K < M owing to one of 3 criteria:
            (1) 'kaiser': variant of the Kaiser-Guttman rule, retaining eigenvalues larger than the median
            (2) 'mcssa': Monte-Carlo SSA (use modes above the 95% quantile from an AR(1) process)
            (3) 'var': first K modes that explain at least var_thresh % of the variance.
        Default is None, which bypasses truncation (K = M)

    var_thresh : float
        variance threshold for reconstruction (only impactful if trunc is set to 'var')

    Returns
    -------
    res : dict containing:

        - eigvals : (M, ) array of eigenvalues

        - eigvecs : (M, M) Matrix of temporal eigenvectors (T-EOFs)

        - PC : (N - M + 1, M) array of principal components (T-PCs)

        - RCmat : (N,  M) array of reconstructed components

        - RCseries : (N,) reconstructed series, with mean and variance restored

        - pctvar: (M, ) array of the fraction of variance (%) associated with each mode

        - eigvals_q : (M, 2) array containting the 5% and 95% quantiles of the Monte-Carlo eigenvalue spectrum [ if nMC >0 ]

        - mode_idx : array of indices of eigenvalues >=eigvals_q

    References
    ----------
    [1]_ Vautard, R., and M. Ghil (1989), Singular spectrum analysis in nonlinear
    dynamics, with applications to paleoclimatic time series, Physica D, 35,
    395–424.

    [2]_ Ghil, M., R. M. Allen, M. D. Dettinger, K. Ide, D. Kondrashov, M. E. Mann,
    A. Robertson, A. Saunders, Y. Tian, F. Varadi, and P. Yiou (2002),
    Advanced spectral methods for climatic time series, Rev. Geophys., 40(1),
    1003–1052, doi:10.1029/2000RG000092.

    [3]_ Allen, M. R., and L. A. Smith (1996), Monte Carlo SSA: Detecting irregular
    oscillations in the presence of coloured noise, J. Clim., 9, 3373–3404.

    [4]_ Schoellhamer, D. H. (2001), Singular spectrum analysis for time series with
    missing data, Geophysical Research Letters, 28(16), 3187–3190, doi:10.1029/2000GL012698.

    See also
    --------
    
    pyleoclim.core.series.Series.ssa : Singular Spectrum Analysis for timeseries objects
    
    pyleoclim.core.ssares.SsaRes.modeplot : plot SSA modes
    
    pyleoclim.core.ssares.SsaRes.screeplot : plot SSA eigenvalue spectrum

    '''

    ys, mu, scale = standardize(y)

    N = len(ys)

    if M == None:
        M=int(N/10)
    c = np.zeros(M)

    for j in range(M):
        prod = ys[0:N - j] * ys[j:N]
        c[j] = sum(prod[~np.isnan(prod)]) / (sum(~np.isnan(prod)) - 1)


    C = toeplitz(c[0:M])  #form correlation matrix

    D, eigvecs = eigh(C) # solve eigendecomposition

    sort_tmp = np.sort(D)
    eigvals = sort_tmp[::-1]
    sortarg = np.argsort(-D)
    eigvecs = eigvecs[:, sortarg]

    # determine principal component time series
    PC = np.zeros((N - M + 1, M))
    PC[:, :] = np.nan
    for k in np.arange(M):
        for i in np.arange(0, N - M + 1):
            #   modify for nan
            prod = ys[i:i + M] * eigvecs[:, k]
            ngood = sum(~np.isnan(prod))
            #   must have at least m*f good points
            if ngood >= M * f:
                PC[i, k] = sum(
                    prod[~np.isnan(prod)]) * M / ngood  # the columns of this matrix are Ak(t), k=1 to M (T-PCs)

    pctvar = eigvals**2/np.sum(eigvals**2)*100 # percent variance

    if nMC > 0: # If Monte-Carlo SSA is requested.
        trunc == 'mcssa'
        noise = ar1_sim(ys, nMC)  # generate MC AR(1) surrogates of y
        eigvals_R = np.zeros((M,nMC)) # define eigenvalue matrix
        lgs = np.arange(-N + 1, N)

        for m in range(nMC):
            xn, _ , _ = standardize(noise[:, m]) # center and standardize
            Gn = np.correlate(xn, xn, "full")
            Gn = Gn / (N - abs(lgs))
            Cn = toeplitz(Gn[N - 1:N - 1 + M])
            eigvals_R[:, m] = np.diag(np.dot(np.dot(np.transpose(eigvecs), Cn), eigvecs))

        eigvals_q = np.empty((M,2))
        eigvals_q[:,0] = np.percentile(eigvals_R, 5, axis=1)
        eigvals_q[:,1] = np.percentile(eigvals_R, 95, axis=1)
        mode_idx = np.where(eigvals>=eigvals_q[:,1])[0]
    else:
        eigvals_q = None


    if trunc is None:
        mode_idx = np.arange(M)
    elif trunc == 'kaiser':
        mval     = np.median(eigvals) # median eigenvalues
        mode_idx = np.where(eigvals>=mval)[0]
    elif trunc == 'var':
        mode_idx = np.arange(np.argwhere(np.cumsum(pctvar)>=var_thresh)[0]+1)
    if nMC == 0 and trunc == 'mcssa':
        raise ValueError('nMC must be larger than 0 to enable MC-SSA truncation')
    elif nMC>0:
       mode_idx = np.where(eigvals>=eigvals_q[:,1])[0]
       
    nmodes = len(mode_idx)

    # compute reconstructed timeseries
    Np = N - M + 1
    RCmat = np.zeros((N, M))

    for im in range(M):
        xdum = np.dot(np.expand_dims(PC[:, im], axis=1), np.expand_dims(eigvecs[0:M, im], axis=0))
        xdum = np.flipud(xdum)

        for n in np.arange(N):
            RCmat[n, im] = np.diagonal(xdum, offset=-(Np - 1 - n)).mean()
        del xdum

    RCmat = scale*RCmat + np.tile(mu, reps=[N, M])  # restore the mean and variance

    #RCseries = scale*RCmat[:,mode_idx].sum(axis=1) + mu

    RCseries = RCmat[:,mode_idx].sum(axis=1)
    if nmodes > 1:
        RCseries -= mu*(nmodes-1)

    # export results
    res = {'eigvals': eigvals, 'eigvecs': eigvecs, 'PC': PC, 'RCseries': RCseries, 'RCmat': RCmat, 'pctvar': pctvar, 'eigvals_q': eigvals_q, 'mode_idx': mode_idx}
    return res



 # -*- coding: utf-8 -*-
"""
Utilities to manipulate LiPD files and automate data transformation whenever possible. 
These functions are used throughout Pyleoclim but are not meant for direct interaction by users.
Also handles integration with the LinkedEarth wiki and the LinkedEarth Ontology.
"""

import lipd as lpd
import numpy as np
import os
import json
import requests
import wget
#from unidecode import unidecode
import string


"""
The following functions handle the LiPD files
"""

def enumerateLipds(lipds):
    """Enumerate the LiPD files loaded in the workspace

    Parameters
    ----------

    lipds : dict
        A dictionary of LiPD files.

    """
    print("Below are the available records")
    lipds_list = [val for val in lipds.keys()]
    for idx, val in enumerate(lipds_list):
        print(idx,': ',val)

def getLipd(lipds):
    """Prompt for a LiPD file

    Ask the user to select a LiPD file from a list
    Use this function in conjunction with enumerateLipds()

    Parameters
    ----------

    lipds : dict
        A dictionary of LiPD files. Can be obtained from
        pyleoclim.readLipd()

    Returns
    -------

    select_lipd : int
        The index of the LiPD file

    """
    enumerateLipds(lipds)
    lipds_list = [val for val in lipds.keys()]
    choice = int(input("Enter the number of the file: "))
    lipd_name = lipds_list[choice]
    select_lipd = lipds[lipd_name]

    return select_lipd


"""
The following functions work at the variables level
"""

def promptForVariable():
    """Prompt for a specific variable

    Ask the user to select the variable they are interested in.
    Use this function in conjunction with readHeaders() or getTSO()

    Returns
    -------

    select_var : int
        The index of the variable

    """
    select_var = int(input("Enter the number of the variable you wish to use: "))
    return select_var

def xAxisTs(timeseries):
    """ Get the x-axis for the timeseries.

    Parameters
    ----------

    timeseries : dict
        a timeseries object

    Returns
    -------

    x_axis : array
        the values for the x-axis representation
    label : string
        returns either "age", "year", or "depth"
    """

    if "depth" in timeseries.keys() and "age" in timeseries.keys() or\
            "depth" in timeseries.keys() and "year" in timeseries.keys():
        print("Both time and depth information available, selecting time")
        if "age" in timeseries.keys() and "year" in timeseries.keys():
            print("Both age and year representation available, selecting age")
            x_axis = timeseries["age"]
            label = "age"
        elif "year" in timeseries.keys():
            x_axis = timeseries["year"]
            label = "year"
        elif "age" in timeseries.keys():
            x_axis = timeseries["age"]
    elif "depth" in timeseries.keys():
        x_axis =  timeseries["depth"]
        label = "depth"
    elif "age" in timeseries.keys():
        x_axis = timeseries["age"]
        label = "age"
    elif "year" in timeseries.keys():
        x_axis = timeseries["year"]
        label = "year"
    else:
        raise KeyError("No age or depth information available")

    return x_axis, label

def checkXaxis(timeseries, x_axis= None):
    """Check that a x-axis is present for the timeseries

    Parameters
    ----------

    timeseries : dict
        a timeseries

    x_axis : string
        the x-axis representation, either depth, age or year

    Returns
    -------

    x : array
        the values for the x-axis representation

    label : string
        returns either "age", "year", or "depth"

    """
    if x_axis is None:
        x, label = xAxisTs(timeseries)
        x = np.array(x, dtype = 'float64')
    elif x_axis == "depth":
        if not "depth" in timeseries.keys():
            raise ValueError("Depth not available for this record")
        else:
            x = np.array(timeseries['depth'], dtype = 'float64')
            label = "depth"
    elif x_axis == "age":
        if not "age" in timeseries.keys():
            raise ValueError("Age not available for this record")
        else:
            x = np.array(timeseries['age'], dtype = 'float64')
            label = "age"
    elif x_axis == "year":
        if not "year" in timeseries.keys():
            raise ValueError("Year not available for this record")
        else:
            x = np.array(timeseries['year'], dtype = 'float64')
            label = "year"
    else:
        raise KeyError("enter either 'depth','age',or 'year'")

    return x, label

def checkTimeAxis(timeseries, x_axis = None):
    """ This function makes sure that time is available for the timeseries

    Parameters
    ----------

    timeseries : dict
        A LiPD timeseries object

    Returns
    -------

    x : array
        the time values for the timeseries

    label : string
        the time representation for the timeseries
    """
    if x_axis is None:
        if not 'age' in timeseries.keys() and not 'year' in timeseries.keys():
            raise KeyError("No time information available")
        elif 'age' in timeseries.keys() and 'year' in timeseries.keys():
            print("Both age and year information are available, using age")
            label = 'age'
        elif 'age' in timeseries.keys():
            label = 'age'
        elif 'year' in timeseries.keys():
            label = 'year'
    elif x_axis == 'age':
        if not 'age' in timeseries.keys():
            raise KeyError('Age is not available for this record')
        else:
            label = 'age'
    elif x_axis == 'year':
        if not 'year' in timeseries.keys():
            raise KeyError('Year is not available for this record')
        else:
            label='year'
    else:
        raise KeyError('Only None, year and age are valid entries for x_axis parameter')

    x = np.array(timeseries[label], dtype = 'float64')

    return x, label

def searchVar(timeseries_list, key, exact = True, override = True):
    """ This function searched for keywords (exact match) for a variable

    Parameters
    ----------

    timeseries_list : list
        A list of available series

    key : list
        A list of keys to search

    exact : bool
        if True, looks for an exact match.

    override : bool
        if True, override the exact match if no match is found

    Returns
    -------

    match : list
        A list of keys for the timeseries that match the selection
        criteria.
    """

    # Make sure thaat the keys are contained in a list
    if type(key) is not list:
       if type(key) is str:
           key = [key]
       else:
           raise TypeError("Key terms should be entered as a list")

    match = []

    if exact == True:
    #Search for exact match with the key
        for keyVal in key:
            for val in timeseries_list.keys():
                ts_temp = timeseries_list[val]
                if "variableName" in ts_temp.keys():
                    name = ts_temp["variableName"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "paleoData_variableName" in ts_temp.keys():
                    name = ts_temp["paleoData_variableName"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "chronData_variableName" in ts_temp.keys():
                    name = ts_temp["chronData_variableName"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "ProxyObservationType" in ts_temp.keys():
                    name = ts_temp["ProxyObservationType"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "paleoData_proxyObservationType" in ts_temp.keys():
                    name = ts_temp["paleoData_proxyObservationType"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "chronData_proxyObservationType" in ts_temp.keys():
                    name = ts_temp["chronData_proxyObservationType"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "InferredVariableType" in ts_temp.keys():
                    name = ts_temp["InferredVariableType"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "paleoData_inferredVariableType" in ts_temp.keys():
                    name = ts_temp["paleoData_inferredVariableType"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
                elif "chronData_inferredVariableType" in ts_temp.keys():
                    name = ts_temp["chronData_inferredVariableType"]
                    if keyVal.lower() == name.lower():
                        match.append(val)
    else:
    # Search for the word in the ley
        for keyVal in key:
            for val in timeseries_list.keys():
                ts_temp = timeseries_list[val]
                if "variableName" in ts_temp.keys():
                    name = ts_temp["variableName"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "paleoData_variableName" in ts_temp.keys():
                    name = ts_temp["paleoData_variableName"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "chronData_variableName" in ts_temp.keys():
                    name = ts_temp["chronData_variableName"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "ProxyObservationType" in ts_temp.keys():
                    name = ts_temp["ProxyObservationType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "paleoData_proxyObservationType" in ts_temp.keys():
                    name = ts_temp["paleoData_proxyObservationType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "chronData_proxyObservationType" in ts_temp.keys():
                    name = ts_temp["chronData_proxyObservationType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "InferredVariableType" in ts_temp.keys():
                    name = ts_temp["InferredVariableType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "paleoData_inferredVariableType" in ts_temp.keys():
                    name = ts_temp["paleoData_inferredVariableType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "chronData_inferredVariableType" in ts_temp.keys():
                    name = ts_temp["chronData_inferredVariableType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)

    # Expand the search if asked
    if not match and exact == True and override == True:
        print("No match found on exact search, running partial match")
        for keyVal in key:
            for val in timeseries_list.keys():
                ts_temp = timeseries_list[val]
                if "variableName" in ts_temp.keys():
                    name = ts_temp["variableName"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "paleoData_variableName" in ts_temp.keys():
                    name = ts_temp["paleoData_variableName"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "chronData_variableName" in ts_temp.keys():
                    name = ts_temp["chronData_variableName"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "ProxyObservationType" in ts_temp.keys():
                    name = ts_temp["ProxyObservationType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "paleoData_proxyObservationType" in ts_temp.keys():
                    name = ts_temp["paleoData_proxyObservationType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "chronData_proxyObservationType" in ts_temp.keys():
                    name = ts_temp["chronData_proxyObservationType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "InferredVariableType" in ts_temp.keys():
                    name = ts_temp["InferredVariableType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "paleoData_inferredVariableType" in ts_temp.keys():
                    name = ts_temp["paleoData_inferredVariableType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)
                elif "chronData_inferredVariableType" in ts_temp.keys():
                    name = ts_temp["chronData_inferredVariableType"]
                    if keyVal.lower() in name.lower():
                        match.append(val)

    # Get the unique entries
    match = list(set(match))

    # Narrow down if more than one match is found by asking the user
    if len(match) > 1:
        print("More than one series match your search criteria")
        for idx, val in enumerate(match):
            print(idx,": ", val)
        choice = int(input("Enter the number for the variable: "))
        match = match[choice]
    elif not match:
        print("No match found.")
        print("Here are the available variables: ")
        v = list(timeseries_list.keys())
        for idx, val in enumerate(v):
            print(idx,": ",val)
        choice = input("Please select the variable you'd like to use or enter to continue: ")
        if not choice:
            match =""
        else:
            choice = int(choice)
            match = v[choice]
    else:
        match = match[0]

    return match

"""
The following functions handle the time series objects
"""
def enumerateTs(timeseries_list):
    """Enumerate the available time series objects

    Parameters
    ----------

    timeseries_list : list
        a  list of available timeseries objects.


    """
    available_y = []
    dataSetName =[]
    at =[]
    for item in timeseries_list:
        if 'dataSetName' in item.keys():
            dataSetName.append(item['dataSetName'])
        else:
            dataSetName.append('NA')
        if 'paleoData_variableName' in item.keys():
            available_y.append(item['paleoData_variableName'])
        elif 'chronData_variableName' in item.keys():
            available_y.append(item['chronData_variableName'])
        else:
            available_y.append('NA')
        if 'archiveType' in item.keys():
            at.append(item['archiveType'])
        else:
            at.append('NA')

    for idx,val in enumerate(available_y):
        print(idx,': ',dataSetName[idx], ': ',at[idx],': ', val)

def getTs(timeseries_list, option = None):
    """Get a specific timeseries object from a dictionary of timeseries

    Parameters
    ----------

    timeseries_list : list
        a  list of available timeseries objects.

    option : string
        An expression to filter the datasets. Uses lipd.filterTs()

    Returns
    -------

    timeseries : single timeseries object or list of timeseries
        A single timeseries object if not optional filter selected or a filtered
        list if optional arguments given
        
    See also
    --------
    
    pyleoclim.utils.lipdutils.enumerateTs : Enumerate the available time series objects

    pyleoclim.utils.lipdutils.promptForVariable : Prompt for a specific variable

    """
    if not option:
        enumerateTs(timeseries_list)
        select_TSO = promptForVariable()
        timeseries = timeseries_list[select_TSO]
    else:
        timeseries = lpd.filterTs(timeseries_list, option)

    return timeseries

"""
Functions to handle data on the LinkedEarth Platform
"""
def LipdToOntology(archiveType):
    """ standardize archiveType

    Transform the archiveType from their LiPD name to their ontology counterpart

    Parameters
    ----------

    archiveType : string
        name of the archiveType from the LiPD file

    Returns
    -------

    archiveType : string
        archiveType according to the ontology

    """
    #Align with the ontology
    if archiveType.lower().replace(" ", "") == "icecore":
        archiveType = 'glacierice'
    elif archiveType.lower().replace(" ", "") == "ice-other":
        archiveType = 'glacierice'
    elif archiveType.lower().replace(" ", "") == 'tree':
        archiveType = 'wood'
    elif archiveType.lower().replace(" ", "") == 'borehole':
        archiveType = 'ice/rock'
    elif archiveType.lower().replace(" ", "") == 'bivalve':
        archiveType = 'molluskshells'

    return archiveType

def timeUnitsCheck(units):
    """ This function attempts to make sense of the time units by checking for equivalence

    Parameters
    ----------

    units : string
        The units string for the timeseries

    Returns
    -------

    unit_group : string
        Whether the units belongs to age_units, kage_units, year_units, ma_units, or undefined
    """

    age_units = ['year B.P.','yr B.P.','yr BP','BP','yrs BP','years B.P.',\
                 'yr. BP','yr. B.P.', 'cal. BP', 'cal B.P.', \
                 'year BP','years BP']
    kage_units = ['kyr BP','kaBP','ka BP','ky','kyr','kyr B.P.', 'ka B.P.', 'ky BP',\
                  'kyrs BP','ky B.P.', 'kyrs B.P.', 'kyBP', 'kyrBP']
    year_units = ['AD','CE','year C.E.','year A.D.', 'year CE','year AD',\
                  'years C.E.','years A.D.','yr CE','yr AD','yr C.E.'\
                  'yr A.D.', 'yrs C.E.', 'yrs A.D.', 'yrs CE', 'yrs AD']
    mage_units = ['my BP', 'myr BP', 'myrs BP', 'ma BP', 'ma',\
                  'my B.P.', 'myr B.P.', 'myrs B.P.', 'ma B.P.']
    undefined = ['years', 'yr','year','yrs']

    if units in age_units:
        unit_group = 'age_units'
    elif units in kage_units:
        unit_group = 'kage_units'
    elif units in year_units:
        unit_group = 'year_units'
    elif units in undefined:
        unit_group = 'undefined'
    elif units in mage_units:
        unit_group = 'ma_units'
    else:
        unit_group = 'unknown'

    return unit_group



def whatArchives(print_response=True):
    """ Get the names for ArchiveType from LinkedEarth Ontology

    Parameters
    ----------

    print_response : bool
        Whether to print the results on the console. Default is True

    Returns
    -------

    res : JSON-object with the request from LinkedEarth wiki api

    """
    url = "http://wiki.linked.earth/store/ds/query"

    query = """PREFIX core: <http://linked.earth/ontology#>
    PREFIX wiki: <http://wiki.linked.earth/Special:URIResolver/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT distinct ?a
    WHERE {
    {
        ?dataset wiki:Property-3AArchiveType ?a.
    }UNION
    {
        ?w core:proxyArchiveType ?t.
        ?t rdfs:label ?a
    }
    }"""

    response = requests.post(url, data = {'query': query})
    res_i = json.loads(response.text)

    if print_response == True:
        print("The following archive types are available on the wiki:")
        for item in res_i['results']['bindings']:
            print ("*" + item['a']['value'])
    res=[]
    for item in res_i['results']['bindings']:
        res.append(item['a']['value'])

    return res

def whatProxyObservations(print_response=True):
    """ Get the names for ProxyObservations from LinkedEarth Ontology

    Parameters
    ----------

    print_response : bool
        Whether to print the results on the console. Default is True

    Returns
    -------

    res : JSON-object with the request from LinkedEarth wiki api
    """

    url = "http://wiki.linked.earth/store/ds/query"

    query = """PREFIX core: <http://linked.earth/ontology#>
    PREFIX wiki: <http://wiki.linked.earth/Special:URIResolver/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT distinct ?a
    WHERE
    {
        ?w core:proxyObservationType ?t.
        ?t rdfs:label ?a
    }"""

    response = requests.post(url, data = {'query': query})
    res_i = json.loads(response.text)

    if print_response==True:
        print("The following proxy observation types are available on the wiki: ")
        for item in res_i['results']['bindings']:
            print (item['a']['value'])

    res=[]
    for item in res_i['results']['bindings']:
        res.append(item['a']['value'])

    return res

def whatProxySensors(print_response=True):
    """ Get the names for ProxySensors from LinkedEarth Ontology

    Parameters
    ----------

    print_response : bool
        Whether to print the results on the console. Default is True

    Returns
    -------

    res : JSON-object with the request from LinkedEarth wiki api
    """
    url = "http://wiki.linked.earth/store/ds/query"

    query = """PREFIX core: <http://linked.earth/ontology#>
    PREFIX wiki: <http://wiki.linked.earth/Special:URIResolver/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT distinct ?a ?b
    WHERE
    {
        ?w core:sensorGenus ?a.
        ?w core:sensorSpecies ?b .

    }"""

    response = requests.post(url, data = {'query': query})
    res_i = json.loads(response.text)

    if print_response == True:
        print("The available sensor genus/species are: ")
        for item in res_i['results']['bindings']:
            print ("*" + 'Genus: '+item['a']['value']+' Species: ' +item['b']['value'])

    res=[]
    for item in res_i['results']['bindings']:
        res.append(item['a']['value'])

    return res

def whatInferredVariables(print_response=True):
    """ Get the names for InferredVariables from LinkedEarth Ontology

    Parameters
    ----------

    print_response : bool
        Whether to print the results on the console. Default is True

    Returns
    -------

    res : JSON-object with the request from LinkedEarth wiki api
    """
    url = "http://wiki.linked.earth/store/ds/query"

    query = """PREFIX core: <http://linked.earth/ontology#>
    PREFIX wiki: <http://wiki.linked.earth/Special:URIResolver/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT distinct ?a
    WHERE
    {
        ?w core:inferredVariableType ?t.
        ?t rdfs:label ?a
    }"""

    response = requests.post(url, data = {'query': query})
    res_i = json.loads(response.text)

    if print_response == True:
        print("The following Inferred Variable types are available on the wiki: ")
        for item in res_i['results']['bindings']:
            print ("*" + item['a']['value'])
            res=[]
    for item in res_i['results']['bindings']:
        res.append(item['a']['value'])
    return res

def whatInterpretations(print_response=True):
    """ Get the names for interpretations from LinkedEarth Ontology

    Parameters
    ----------

    print_response : bool
        Whether to print the results on the console. Default is True

    Returns
    -------

    res : JSON-object with the request from LinkedEarth wiki api
    """

    url = "http://wiki.linked.earth/store/ds/query"

    query = """PREFIX core: <http://linked.earth/ontology#>
    PREFIX wiki: <http://wiki.linked.earth/Special:URIResolver/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT distinct ?a ?b
    WHERE
    {
        ?w core:name ?a.
        ?w core:detail ?b .

    }"""

    response = requests.post(url, data = {'query': query})
    res_i = json.loads(response.text)

    if print_response==True:
        print("The following interpretation are available on the wiki: ")
        for item in res_i['results']['bindings']:
            print ("*" + 'Name: '+item['a']['value']+' Detail: ' +item['b']['value'])
    res=[]
    for item in res_i['results']['bindings']:
        res.append(item['a']['value'])
    return res

def queryLinkedEarth(archiveType=[ ], proxyObsType=[ ], infVarType = [ ], sensorGenus=[ ],
                    sensorSpecies=[ ], interpName =[ ], interpDetail =[ ], ageUnits = [ ],
                    ageBound = [ ], ageBoundType = [ ], recordLength = [ ], resolution = [ ],
                    lat = [ ], lon = [ ], alt = [ ], print_response = True, download_lipd = True,
                    download_folder = 'default'):
    """ This function allows to query the LinkedEarth wiki for records.

    This function allows to query the LinkedEarth wiki for specific catagories.
    If you have more than one keyword per catagory, enter them in a list. If you don't
    wish to use a particular terms, leave a blank in-between the brackets.

    Parameters
    ----------

    archiveType : list of strings
        The type of archive (enter all query terms, separated by a comma)

    proxyObsType : list of strings
        The type of proxy observation (enter all query terms, separated by a comma)

    infVarType : list of strings
        The type of inferred variable (enter all query terms, separated by a comma)

    sensorGenus : list of strings
        The Genus of the sensor (enter all query terms, separated by a comma)

    sensorSpecies : list of strings
        The Species of the sensor (enter all query terms, separated by a comma)

    interpName : list of strings
        The name of the interpretation (enter all query terms, separated by a comma)

    interpDetail : list of strings
        The detail of the interpretation (enter all query terms, separated by a comma)

    ageUnits : list of strings
        The units of in which the age (year) is expressed in.
        Warning: Separate each query if need to run across multiple age queries (i.e., yr B.P. vs kyr B.P.). If the units are different but the meaning is the same (e.g., yr B.P. vs yr BP, enter all search terms separated by a comma).

    ageBound : list of floats
        Enter the minimum and maximum age value to search for.
        Warning: You MUST enter a minimum AND maximum value. If you wish to perform a query such as "all ages before 2000 A.D.", enter a minimum value of -99999 to cover all bases.

    ageBoundType : list of strings
        The type of querying to perform. Possible values include: "any", "entire", and "entirely".
        - any: Overlap any portions of matching datasets (default)
        - entirely: are entirely overlapped by matching datasets
        - entire: overlap entire matching datasets but dataset can be shorter than the bounds

    recordLength : list of floats
        The minimum length the record needs to have while matching the ageBound criteria. For instance, "look for all records between 3000 and 6000 year BP with a record length of at least 1500 year".

    resolution : list of floats
        The maximum resolution of the resord. Resolution has the same units as age/year. For instance, "look for all records with a resolution of at least 100 years".
        Warning: Resolution applies to specific variables rather than an entire dataset. Imagine the case where some measurements are made every cm while others are made every 5cm. If you require a specific variable to have the needed resolution, make sure that either the proxyObservationType, inferredVariableType, and/or Interpretation fields are completed.

    lat : list of floats
        The minimum and maximum latitude. South is expressed with negative numbers.
        Warning: You MUST enter a minimum AND maximum value. If you wish to perform a query looking for records from the Northern Hemisphere, enter [0,90].

    lon : list of floats
        The minimum and maximum longitude. West is expressed with negative numbers.
        Warning: You MUST enter a minimum AND a maximum value. If you wish to perform a query looking for records from the Western Hemisphere, enter [-180,0].

    alt : list of floats
        The minimum and maximum altitude. Depth below sea level is expressed as negative numbers.
        Warning: You MUST enter a minimum AND a maximum value. If you wish to perform a query looking for records below a certain depth (e.g., 500), enter [-99999,-500].

    print_response : bool
        If True, prints the URLs to the matching LiPD files

    download_lipd : bool
        If True, download the matching LiPD files

    download_folder : string
        Location to download the LiPD files. If "default", will download in the current directory.

    Returns
    -------

    res : the response to the query

    """
    # Perform a lot of checks
    if len(ageBound)==1:
        raise ValueError("You need to provide a minimum and maximum boundary.")

    if ageBound and not ageUnits:
        raise ValueError("When providing age limits, you must also enter the units")

    if recordLength and not ageUnits:
        raise ValueError("When providing a record length, you must also enter the units")

    if ageBound and ageBound[0]>ageBound[1]:
        ageBound = [ageBound[1],ageBound[0]]

    if len(ageBoundType)>1:
        raise ValueError("Only one search possible at a time.")
        if ageBoundType not in ["any","entirely","entire"]:
            raise ValueError("ageBoundType is not recognized")

    if recordLength and ageBound and recordLength[0] > (ageBound[1]-ageBound[0]):
        raise ValueError("The required recordLength is greater than the provided age bounds")

    if len(resolution)>1:
        raise ValueError("You can only search for a maximum resolution one at a time.")

    if len(lat)==1:
        raise ValueError("Please enter a lower AND upper boundary for the latitude search")

    if lat and lat[1]<lat[0]:
        lat = [lat[1],lat[0]]

    if len(lon)==1:
        raise ValueError("Please enter a lower AND upper boundary for the longitude search")

    if lon and lon[1]<lon[0]:
        lon = [lon[1],lon[0]]

    if len(alt)==1:
        raise ValueError("Please enter a lower AND upper boundary for the altitude search")

    if alt and alt[1]<alt[0]:
        alt = [alt[1],alt[0]]

    # Perform the query
    url = "http://wiki.linked.earth/store/ds/query"

    query = """PREFIX core: <http://linked.earth/ontology#>
    PREFIX wiki: <http://wiki.linked.earth/Special:URIResolver/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

    SELECT  distinct ?dataset ?dataset_label
    WHERE {
    ?dataset rdfs:label ?dataset_label
    """

    ### Look for data field
    dataQ=""
    if archiveType or proxyObsType or infVarType or sensorGenus or sensorSpecies or interpName or interpDetail or ageUnits or ageBound or recordLength or resolution:
        dataQ = "?dataset core:includesChronData|core:includesPaleoData ?data."

    ### Look for variable
    ## measuredVariable
    measuredVarQ=""
    if proxyObsType or archiveType or sensorGenus or sensorSpecies or interpName or interpDetail or resolution:
        measuredVarQ = "?data core:foundInMeasurementTable / core:includesVariable ?v."

    ## InferredVar
    inferredVarQ=""
    if infVarType or interpName or interpDetail or resolution:
        inferredVarQ = "?data core:foundInMeasurementTable / core:includesVariable ?v1."

    ### Archive Query
    archiveTypeQ=""
    if len(archiveType)>0:
        #add values for the archiveType
        query += "VALUES ?a {"
        for item in archiveType:
            query +="\""+item+"\" "
        query += "}\n"
        # Create the query
        archiveTypeQ = """
    #Archive Type query
    {
        ?dataset wiki:Property-3AArchiveType ?a.
    }UNION
    {
        ?p core:proxyArchiveType / rdfs:label ?a.
    }"""

    ### ProxyObservationQuery
    proxyObsTypeQ=""
    if len(proxyObsType)>0:
       #  add values for the proxyObservationType
       query+="VALUES ?b {"
       for item in proxyObsType:
           query += "\""+item+"\""
       query += "}\n"
       # Create the query
       proxyObsTypeQ="?v core:proxyObservationType/rdfs:label ?b."

    ### InferredVariableQuery
    infVarTypeQ=""
    if len(infVarType)>0:
        query+="VALUES ?c {"
        for item in infVarType:
            query+="\""+item+"\""
        query+="}\n"
        # create the query
        infVarTypeQ="""
    ?v1 core:inferredVariableType ?t.
    ?t rdfs:label ?c.
    """
    ### ProxySensorQuery
    sensorQ=""
    if len(sensorGenus)>0 or len(sensorSpecies)>0:
        sensorQ="""
    ?p core:proxySensorType ?sensor.
    """
    ## Genus query
    genusQ=""
    if len(sensorGenus)>0:
        query+="VALUES ?genus {"
        for item in sensorGenus:
            query+="\""+item+"\""
        query+="}\n"
        # create the query
        genusQ = "?sensor core:sensorGenus ?genus."

    ## Species query
    speciesQ=""
    if len(sensorSpecies)>0:
        query+=  "VALUES ?species {"
        for item in sensorSpecies:
            query+="\""+item+"\""
        query+="}\n"
        #Create the query
        speciesQ = "?sensor core:sensorSpecies ?species."

    ### Proxy system query
    proxySystemQ = ""
    if len(archiveType)>0 or len(sensorGenus)>0 or len(sensorSpecies)>0:
        proxySystemQ="?v ?proxySystem ?p."

    ### Deal with interpretation
    ## Make sure there is an interpretation to begin with
    interpQ = ""
    if len(interpName)>0 or len(interpDetail)>0:
        interpQ= """
    {?v1 core:interpretedAs ?interpretation}
    UNION
    {?v core:interpretedAs ?interpretation}
    """

    ## Name
    interpNameQ=""
    if len(interpName)>0:
        query+= "VALUES ?intName {"
        for item in interpName:
            query+="\""+item+"\""
        query+=  "}\n"
        #Create the query
        interpNameQ = "?interpretation core:name ?intName."

    ## detail
    interpDetailQ = ""
    if len(interpDetail)>0:
        query+= "VALUES ?intDetail {"
        for item in interpDetail:
            query+="\""+item+"\""
        query+="}\n"
        #Create the query
        interpDetailQ = "?interpretation core:detail ?intDetail."

    ### Age
    ## Units
    ageUnitsQ = ""
    if len(ageUnits)>0:
        query+= "VALUES ?units {"
        for item in ageUnits:
            query+="\""+item+"\""
        query+="}\n"
        query+="""VALUES ?ageOrYear{"Age" "Year"}\n"""
        # create the query
        ageUnitsQ ="""
    ?data core:foundInMeasurementTable / core:includesVariable ?v2.
    ?v2 core:inferredVariableType ?aoy.
    ?aoy rdfs:label ?ageOrYear.
    ?v2 core:hasUnits ?units .
    """
    ## Minimum and maximum
    ageQ = ""
    if ageBoundType[0] == "entirely":
        if len(ageBound)>0 and len(recordLength)>0:
            ageQ="""
    ?v2 core:hasMinValue ?e1.
    ?v2 core:hasMaxValue ?e2.
    filter(?e1<=""" +str(ageBound[0])+ """&& ?e2>="""+str(ageBound[1])+""" && abs(?e1-?e2)>="""+str(recordLength[0])+""").
    """
        elif len(ageBound)>0 and len(recordLength)==0:
            ageQ="""
    ?v2 core:hasMinValue ?e1.
    ?v2 core:hasMaxValue ?e2.
    filter(?e1<=""" +str(ageBound[0])+ """&& ?e2>="""+str(ageBound[1])+""").
    """
    elif ageBoundType[0] == "entire":
        if len(ageBound)>0 and len(recordLength)>0:
            ageQ="""
    ?v2 core:hasMinValue ?e1.
    ?v2 core:hasMaxValue ?e2.
    filter(?e1>=""" +str(ageBound[0])+ """&& ?e2<="""+str(ageBound[1])+""" && abs(?e1-?e2)>="""+str(recordLength[0])+""").
    """
        elif len(ageBound)>0 and len(recordLength)==0:
            ageQ="""
    ?v2 core:hasMinValue ?e1.
    ?v2 core:hasMaxValue ?e2.
    filter(?e1>=""" +str(ageBound[0])+ """&& ?e2<="""+str(ageBound[1])+""").
    """
    elif ageBoundType[0] == "any":
        if len(ageBound)>0 and len(recordLength)>0:
            ageQ="""
    ?v2 core:hasMinValue ?e1.
    filter(?e1<=""" +str(ageBound[1])+ """ && abs(?e1-"""+str(ageBound[1])+""")>="""+str(recordLength[0])+""").
    """
        elif len(ageBound)>0 and len(recordLength)==0:
            ageQ="""
    ?v2 core:hasMinValue ?e1.
    filter(?e1<=""" +str(ageBound[1])+ """).
    """

    ### Resolution
    resQ=""
    if len(resolution)>0:
        resQ = """
    {
    ?v core:hasResolution/(core:hasMeanValue |core:hasMedianValue) ?resValue.
    filter (xsd:float(?resValue)<100)
    }
    UNION
    {
    ?v1 core:hasResolution/(core:hasMeanValue |core:hasMedianValue) ?resValue1.
    filter (xsd:float(?resValue1)<"""+str(resolution[0])+""")
    }
    """

    ### Location
    locQ=""
    if lon or lat or alt:
           locQ = "?dataset core:collectedFrom ?z."

    ## Altitude
    latQ=""
    if len(lat)>0:
        latQ="""
    ?z <http://www.w3.org/2003/01/geo/wgs84_pos#lat> ?lat.
    filter(xsd:float(?lat)<"""+str(lat[1])+""" && xsd:float(?lat)>"""+str(lat[0])+""").
    """

    ##Longitude
    lonQ=""
    if len(lon)>0:
        lonQ = """
    ?z <http://www.w3.org/2003/01/geo/wgs84_pos#long> ?long.
    filter(xsd:float(?long)<"""+str(lon[1])+""" && xsd:float(?long)>"""+str(lon[0])+""").
    """

    ## Altitude
    altQ=""
    if len(alt)>0:
        altQ="""
    ?z <http://www.w3.org/2003/01/geo/wgs84_pos#alt> ?alt.
    filter(xsd:float(?alt)<"""+str(alt[1])+""" && xsd:float(?alt)>"""+str(alt[0])+""").
    """

    query += """
    ?dataset a core:Dataset.
    """+dataQ+"""
    """+measuredVarQ+"""
    # By proxyObservationType
    """+proxyObsTypeQ+"""
    """+inferredVarQ+"""
    # By InferredVariableType
    """+infVarTypeQ+"""
    # Look for the proxy system model: needed for sensor and archive queries
    """+proxySystemQ+"""
    # Sensor query
    """+sensorQ+"""
    """+genusQ+"""
    """+speciesQ+"""
    # Archive query (looks in both places)
    """+archiveTypeQ+"""
    # Interpretation query
    """+interpQ+"""
    """+interpNameQ+"""
    """+interpDetailQ+"""
    # Age Query
    """+ageUnitsQ+"""
    """+ageQ+"""
    # Location Query
    """+locQ+"""
    #Latitude
    """+latQ+"""
    #Longitude
    """+lonQ+"""
    #Altitude
    """+altQ+"""
    #Resolution Query
    """+resQ+"""
    }"""

    #print(query)
    response = requests.post(url, data = {'query': query})
    res = json.loads(response.text)
    if print_response == True:
        for item in res['results']['bindings']:
            print (item['dataset']['value'])

    #download files
    if download_lipd == True:
        for item in res['results']['bindings']:
            dataset = (item['dataset_label']['value'])
            download_url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid='+dataset
            if download_folder == 'default':
                path = os.getcwd()+'/'
            else:
                if download_folder[-1] == '/':
                    path = download_folder
                    wget.download(download_url, path)
                else:
                    path = download_folder+'/'
            if os.path.exists(path) == False:
                os.mkdir(path)
            wget.download(download_url, path+'/'+dataset+'.lpd')

    return res

def pre_process_list(list_str):
    """ Pre-process a series of strings for capitalized letters, space, and punctuation

    Parameters
    ----------

    list_str : list
        A list of strings from which to strip capitals, spaces, and other characters

    Returns
    -------

    res : list
        A list of strings with capitalization, spaces, and punctuation removed
    """
    res=[]
    for item in list_str:
        res.append(pre_process_str(item))
    return res

def similar_string(list_str, search):
    """ Returns a list of indices for strings with similar values

    Parameters
    ----------

    list_str : list
        A list of strings

    search : str
        A keyword search

    Returns
    -------

    indices: list
        A list of indices with similar value as the keyword
    """
    #exact matches
    indices = [i for i, x in enumerate(list_str) if x == search]
    # proximity matches
    return indices

def pre_process_str(word):
    """Pre-process a string for capitalized letters, space, and punctuation

    Parameters
    ----------

    string : str
        A string from which to strip capitals, spaces, and other characters

    Returns
    -------
    res : str
        A string with capitalization, spaces, and punctuation removed
    """
    d=word.replace(" ","").lower()
    stopset=list(string.punctuation)
    res="".join([i for i in d if i not in stopset])
    return res

"""
Deal with models
"""
def isModel(csvName, lipd):
    """Check for the presence of a model in the same object as the measurement table

    Parameters
    ----------

    csvName : string
        The name of the csv file corresponding to the measurement table

    lipd : dict
        A LiPD object

    Returns
    -------

    model : list
        List of models already available

    dataObject : string
        The name of the paleoData or ChronData
        object in which the model(s) are stored
    """
    csvNameSplit = csvName.split('.')
    for val in csvNameSplit:
        if "chron" in val or "paleo" in val:
            tableName = val

    if tableName[0] == 'c':
        objectName = 'chron'+tableName.split('chron')[1][0]
        dataObject = lipd["chronData"][objectName]
    elif tableName[0] == 'p':
        objectName = 'paleo'+tableName.split('paleo')[1][0]
        dataObject = lipd["paleoData"][objectName]
    else:
        raise KeyError("Key name should only include 'chron' or 'paleo'")

    if "model" in dataObject.keys():
        model_list = dataObject["model"]
        model = list(model_list.keys())
    else:
        model=[]

    return model, objectName


def modelNumber(model):
    """Assign a new or existing model number

    Parameters
    ----------

    model : list
        List of possible model number. Obtained from isModel

    Returns
    -------

    modelNum : int
        The number of the model
    """
    if model:
        print("There is " + str(len(model)) + " model(s) already available.")
        print("creating a new model...")
        modelNum = len(model)
        print("Your new model number is "+ str(modelNum))
    else:
        print("No previous model available. Creating a new model...")
        modelNum = 0
        print("Your model number is "+ str(modelNum))

    return modelNum

"""
Get entire tables
"""

def isMeasurement(csv_dict):
    """ Check whether measurement tables are available

    Parameters
    ----------

    csv_dict : dict
        Dictionary of available csv

    Returns
    -------

    paleoMeasurementTables : list
        List of available paleoMeasurementTables
    chronMeasurementTables : list
        List of available chronMeasurementTables
    """
    chronMeasurementTables = []
    paleoMeasurementTables =[]

    for val in csv_dict.keys():
        if "measurement" in val and "chron" in val:
            chronMeasurementTables.append(val)
        if "measurement" in val and "paleo" in val:
            paleoMeasurementTables.append(val)

    return chronMeasurementTables, paleoMeasurementTables

def whichMeasurement(measurementTableList):
    """Select a measurement table from a list

    Use in conjunction with the function isMeasurement

    Parameters
    ----------

    measurementTableList : list
        List of measurement tables contained in the LiPD file. Output from the isMeasurement function

    Returns
    -------

    csvName : string
        the name of the csv file

    """
    if len(measurementTableList)>1:
        print("More than one table is available.")
        for idx, val in enumerate(measurementTableList):
            print(idx, ": ", val)
        csvName = measurementTableList[int(input("Which one would you like to use? "))]
    else:
        csvName = measurementTableList[0]

    return csvName

def getMeasurement(csvName, lipd):
    """Extract the dictionary corresponding to the measurement table

    Parameters
    ----------

    csvName : string
        The name of the csv file
    lipd : dict
        The LiPD object from which to extract the data

    Returns
    -------

    ts_list : dict
        A dictionary containing data and metadata for each column in the
        csv file.

    """
    csvNameSplit = csvName.split('.')
    for val in csvNameSplit:
        if "chron" in val or "paleo" in val:
            tableName = val

    if tableName[0] == 'c':
        objectName = 'chron'+tableName.split('chron')[1][0]
        ts_list = lipd["chronData"][objectName]["measurementTable"][tableName]["columns"]
    elif tableName[0] == 'p':
        objectName = 'paleo'+tableName.split('paleo')[1][0]
        ts_list = lipd["paleoData"][objectName]["measurementTable"][tableName]["columns"]
    else:
        raise KeyError("Key name should only include 'chron' or 'paleo'")

    return ts_list

"""
Deal with ensembles
"""

def isEnsemble(csv_dict):
    """ Check whether ensembles are available

    Parameters
    ----------

    csv_dict : dict
        Dictionary of available csv

    Returns
    -------

    paleoEnsembleTables : list
        List of available paleoEnsembleTables
    chronEnsembleTables : list
        List of availale chronEnsemble Tables

    """
    chronEnsembleTables =[]
    paleoEnsembleTables =[]
    for val in csv_dict.keys():
        if "ensemble" in val and "chron" in val:
            chronEnsembleTables.append(val)
        elif "ensemble" in val and "paleo" in val:
            paleoEnsembleTables.append(val)

    return chronEnsembleTables, paleoEnsembleTables

def whichEnsemble(ensembleTableList):
    """Select an ensemble table from a list

    Use in conjunction with the function isMeasurement

    Parameters
    ----------

    measurementTableList : list
        List of measurement tables contained in the LiPD file. Output from the isMeasurement function
    csv_list : list
        Dictionary of available csv

    Returns
    -------

    csvName : string
        the name of the csv file

    """
    if len(ensembleTableList)>1:
        print("More than one table is available.")
        for idx, val in enumerate(ensembleTableList):
            print(idx, ": ", val)
        csvName = ensembleTableList[int(input("Which one would you like to use? "))]
    else:
        csvName = ensembleTableList[0]

    return csvName

def getEnsemble(csv_dict, csvName):
    """ Extracts the ensemble values and depth vector from the dictionary and
    returns them into two numpy arrays.

    Parameters
    ----------

    csv_dict : dict
        dictionary containing the availableTables

    csvName : str
        Name of the csv

    Returns
    -------

    depth : array
        Vector of depth

    ensembleValues : array
        The matrix of Ensemble values
    """
    ensemble_dict=csv_dict[csvName]
    ensembleValues=[]
    for val in ensemble_dict.keys():
        if 'depth' in val:
            depth = ensemble_dict[val]["values"]
        else:
            ensembleValues.append(ensemble_dict[val]["values"])
    ensembleValues= np.transpose(np.array(ensembleValues))

    return depth, ensembleValues

def mapAgeEnsembleToPaleoData(ensembleValues, depthEnsemble, depthPaleo):
    """ Map the depth for the ensemble age values to the paleo depth

    Parameters
    ----------

    ensembleValues : array
        A matrix of possible age models. Realizations
        should be stored in columns
    depthEnsemble : array
        A vector of depth. The vector should have the same
        length as the number of rows in the ensembleValues
    depthPaleo : array
        A vector corresponding to the depth at which there
        are paleodata information

    Returns
    -------

    ensembleValuesToPaleo : array
        A matrix of age ensemble on the PaleoData scale

    """
    if len(depthEnsemble)!=np.shape(ensembleValues)[0]:
        raise ValueError("Depth and age need to have the same length")

    #Make sure that numpy arrays were given
    ensembleValues=np.array(ensembleValues)
    depthEnsemble=np.array(depthEnsemble)
    depthPaleo = np.array(depthPaleo)

    #Interpolate
    ensembleValuesToPaleo = np.zeros((len(depthPaleo),np.shape(ensembleValues)[1])) #placeholder
    for i in np.arange(0,np.shape(ensembleValues)[1]):
        ensembleValuesToPaleo[:,i]=np.interp(depthPaleo,depthEnsemble,ensembleValues[:,i])

    return ensembleValuesToPaleo

def gen_dict_extract(key, var):
    '''Recursively searches for all the values in nested dictionaries corresponding
    to a particular key

    Parameters
    ----------

    key : str
        The key to search for
    var : dict
        The dictionary to search

    '''
    if hasattr(var,'items'):
        for k, v in var.items():
            if k == key:
                yield v
            if isinstance(v, dict):
                for result in gen_dict_extract(key, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in gen_dict_extract(key, d):
                        yield result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Basic functionalities to clean timeseries data prior to analysis
"""

import numpy as np
from typing import OrderedDict

#APIs

__all__ = [
    'clean_ts',
    'dropna',
    'sort_ts',
    'is_evenly_spaced',
    'reduce_duplicated_timestamps',
]


def clean_ts(ys, ts, verbose=False):
    ''' Cleaning the timeseries

    Delete the NaNs in the time series and sort it with time axis ascending,
    duplicate timestamps will be reduced by averaging the values.

    Parameters
    ----------
    ys : array
        A time series, NaNs allowed
    ts : array
        The time axis of the time series, NaNs allowed
    verbose : bool
        If True, will print a warning message

    Returns
    -------
    ys : array
        The time series without nans
    ts : array
        The time axis of the time series without nans

    See also
    --------

    pyleoclim.utils.tsbase.dropna : Drop NaN values

    pyleoclim.utils.tsbase.sort_ts : Sort timeseries

    pyleoclim.utils.tsbase.reduce_duplicated_timestamps : Consolidate duplicated timestamps

    '''
    ys, ts = dropna(ys, ts, verbose=verbose)
    ys, ts = sort_ts(ys, ts, verbose=verbose)
    ys, ts = reduce_duplicated_timestamps(ys, ts, verbose=verbose)

    return ys, ts


def dropna(ys, ts, verbose=False):
    '''Drop NaN values
    
    Remove entries of ys or ts that bear NaNs

    Parameters
    ----------
    ys : array
        A time series, NaNs allowed
    ts : array
        The time axis of the time series, NaNs allowed
    verbose : bool
        If True, will print a warning message

    Returns
    -------
    ys : array
        The time series without nans
    ts : array
        The time axis of the time series without nans

    '''
    ys = np.asarray(ys, dtype=float)
    ts = np.asarray(ts, dtype=float)
    assert ys.size == ts.size, 'The size of time axis and data value should be equal!'

    ys_tmp = np.copy(ys)
    ys = ys[~np.isnan(ys_tmp)]
    ts = ts[~np.isnan(ys_tmp)]
    ts_tmp = np.copy(ts)
    ys = ys[~np.isnan(ts_tmp)]
    ts = ts[~np.isnan(ts_tmp)]

    if verbose and any(np.isnan(ys_tmp)):
        print('NaNs have been detected and dropped.')

    return ys, ts

def sort_ts(ys, ts, verbose=False):
    ''' Sort timeseries
    
    Sort ts values in ascending order

    Parameters
    ----------
    ys : array
        Dependent variable
    ts : array
        Independent variable
    verbose : bool
        If True, will print a warning message

    Returns
    -------
    ys : array
        Dependent variable
    ts : array
        Independent variable, sorted in ascending order

    '''
    ys = np.asarray(ys, dtype=float)
    ts = np.asarray(ts, dtype=float)
    assert ys.size == ts.size, 'time and value arrays must be of equal length'

    # sort the time series so that the time axis will be ascending
    dt = np.median(np.diff(ts))
    if dt < 0:
        sort_ind = np.argsort(ts)
        ys = ys[sort_ind]
        ts = ts[sort_ind]
        if verbose:
            print('The time axis has been adjusted to be prograde')

    return ys, ts

def reduce_duplicated_timestamps(ys, ts, verbose=False):
    ''' Consolidate duplicated timestamps
    
    Reduce duplicated timestamps in a timeseries by averaging the values

    Parameters
    ----------
    ys : array
        Dependent variable
    ts : array
        Independent variable
    verbose : bool
        If True, will print a warning message

    Returns
    -------
    ys : array
        Dependent variable
    ts : array
        Independent variable, with duplicated timestamps reduced by averaging the values

    '''
    ys = np.asarray(ys, dtype=float)
    ts = np.asarray(ts, dtype=float)
    assert ys.size == ts.size, 'The size of time axis and data value should be equal!'

    if len(ts) != len(set(ts)):
        value = OrderedDict()
        for t, y in zip(ts, ys):
            if t not in value:
                value[t] = [y]
            else:
                value[t].append(y)

        ts = []
        ys = []
        for k, v in value.items():
            ts.append(k)
            ys.append(np.mean(v))

        ts = np.array(ts)
        ys = np.array(ys)

        if verbose:
            print('Duplicate timestamps have been combined by averaging values.')
    return ys, ts

def is_evenly_spaced(ts, tol=1e-4):
    ''' Check if a time axis is evenly spaced, within a given tolerance

    Parameters
    ----------

    ts : array
        The time axis of a time series
        
    tol : float64
        Numerical tolerance for the relative difference

    Returns
    -------

    check : bool
        True - evenly spaced; False - unevenly spaced.

    '''
    if ts is None:
        check = True
    else:
        dts = np.diff(ts)
        dt_mean = dts.mean()   
        check = all(np.abs((dt - dt_mean)/dt_mean) < tol for dt in np.diff(ts)) # compare relative spacing to the mean
        
    return check





''' Tests for pyleoclim.core.ui.SSARes

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''
#import pandas as pd

import pytest
import pyleoclim as pyleo
import numpy as np


# a collection of useful functions

def gen_ts(model, t=None, nt=1000, alpha=1.0):
    'wrapper for gen_ts in pyleoclim'

    t, v = pyleo.utils.gen_ts(model=model, nt=nt, alpha=alpha, t=t)
    ts = pyleo.Series(t, v)
    return ts

# Tests below
      
class TestUiSSAResScreeplot:
    ''' Tests for SSARes.screeplot()
    '''

    def test_plot_t0(self):
        ''' Test SSARes.screeplot with non-default parameters
         (default is already taken care of in TestUISeriesSsa)
        '''
        nt = 500
        t = np.arange(nt)
        cn = gen_ts(model='colored_noise', t=t, alpha=1.0)

        cn_ssa = cn.ssa()
        fig, ax = cn_ssa.screeplot(title="Non default title")
        pyleo.closefig(fig)
        
    def test_plot_t1(self):
        ''' Test SSARes.screeplot with MC-SSA 
        '''
        nt = 500
        t  = np.arange(nt)
        cn = gen_ts(model='colored_noise', t=t, alpha=1.0)

        cn_ssa = cn.ssa(nMC=200)
        fig, ax = cn_ssa.screeplot(title="MC-SSA scree plot")
        pyleo.closefig(fig)
        
        
class TestUiSSAResModeplot:
    ''' Tests for SSARes.modeplot()
    '''
    @pytest.mark.parametrize('spec_method', ['mtm', 'welch', 'periodogram'])
    def test_plot_t0(self, spec_method):
        ''' Test SSARes.modeplot with 3 spectral methods
        '''
        nt = 500
        t  = np.arange(nt)
        cn = gen_ts(model='colored_noise', t=t, alpha=1.0)

        cn_ssa = cn.ssa()
        fig, ax = cn_ssa.modeplot(spec_method=spec_method)
        pyleo.closefig(fig)
        
    def test_plot_t1(self):
        ''' Test SSARes.modeplot with nondefault mode index
        '''
        nt = 500
        t = np.arange(nt)
        cn = gen_ts(model='colored_noise', t=t, alpha=1.0)

        cn_ssa = cn.ssa()
        fig, ax = cn_ssa.modeplot(index=4)
        pyleo.closefig(fig)
        


from ..utils import plotting, lipdutils
from ..utils import wavelet as waveutils
from ..utils import spectral as specutils

#from ..core.multiplepsd import *
#import ..core.multiplepsd


import matplotlib.pyplot as plt
import numpy as np
from tabulate import tabulate
from copy import deepcopy
from tqdm import tqdm
import warnings

from matplotlib.ticker import ScalarFormatter, FormatStrFormatter


def infer_period_unit_from_time_unit(time_unit):
    ''' infer a period unit based on the given time unit

    '''
    if time_unit is None:
        period_unit = None
    else:
        unit_group = lipdutils.timeUnitsCheck(time_unit)
        if unit_group != 'unknown':
            if unit_group == 'kage_units':
                period_unit = 'kyrs'
            else:
                period_unit = 'yrs'
        else:
            if time_unit[-1] == 's':
                period_unit = time_unit
            else:
                period_unit = f'{time_unit}s'

    return period_unit

class PSD:
    '''The PSD (Power spectral density) class is intended for conveniently manipulating
    the result of spectral methods, including performing significance tests,
    estimating scaling coefficients, and plotting. 

    See examples in pyleoclim.core.series.Series.spectral to see how to create and manipulate these objects

    Parameters
    ----------

    frequency : numpy.array, list, or float

        One or more frequencies in power spectrum

    amplitude : numpy.array, list, or float

        The amplitude at each (frequency, time) point;
        note the dimension is assumed to be (frequency, time)

    label : str, optional

        Descriptor of the PSD.
        Default is None

    timeseries : pyleoclim.Series, optional

        Default is None

    plot_kwargs : dict, optional

        Plotting arguments for seaborn histplot: https://seaborn.pydata.org/generated/seaborn.histplot.html.
        Default is None

    spec_method : str, optional

        The name of the spectral method to be applied on the timeseries
        Default is None

    spec_args : dict, optional

        Arguments for wavelet analysis ('freq', 'scale', 'mother', 'param')
        Default is None

    signif_qs : pyleoclim.MultipleScalogram, optional

        Pyleoclim MultipleScalogram object containing the quantiles qs of the surrogate scalogram distribution.
        Default is None

    signif_method : str, optional

        The method used to obtain the significance level.
        Default is None

    period_unit : str, optional

        Unit of time.
        Default is None

    beta_est_res : list or numpy.array, optional

        Results of the beta estimation calculation.
        Default is None.

    See also
    --------

    pyleoclim.core.series.Series.spectral : Spectral analysis

    pyleoclim.core.scalograms.Scalogram :  Scalogram object

    pyleoclim.core.scalograms.MultipleScalogram : Object storing multiple scalogram objects

    pyleoclim.core.psds.MultiplePSD : Object storing several PSDs from different Series or ensemble members in an age model



    '''
    def __init__(self, frequency, amplitude, label=None, timeseries=None, plot_kwargs=None,
                 spec_method=None, spec_args=None, signif_qs=None, signif_method=None, period_unit=None,
                 beta_est_res=None):

        self.frequency = np.array(frequency)
        self.amplitude = np.array(amplitude)
        self.label = label
        self.timeseries = timeseries
        self.spec_method = spec_method
        if spec_args is not None:
            if 'freq' in spec_args.keys():
                spec_args['freq'] = np.array(spec_args['freq'])
        self.spec_args = spec_args
        self.signif_qs = signif_qs
        self.signif_method = signif_method
        self.plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
        if beta_est_res is None:
            self.beta_est_res = beta_est_res
        else:
            self.beta_est_res = np.array(beta_est_res)
        if period_unit is not None:
            self.period_unit = period_unit
        elif timeseries is not None:
            self.period_unit = infer_period_unit_from_time_unit(timeseries.time_unit)
        else:
            self.period_unit = None

    def copy(self):
        '''Copy object
        '''
        return deepcopy(self)

    def __str__(self):
        table = {
            'Frequency': self.frequency,
            'Amplitude': self.amplitude,
        }

        msg = print(tabulate(table, headers='keys'))
        return f'Length: {np.size(self.frequency)}'

    def signif_test(self, method='ar1sim', number=None, seed=None, qs=[0.95],
                    settings=None, scalogram = None):
        '''


        Parameters
        ----------

        number : int, optional

            Number of surrogate series to generate for significance testing. The default is None.

        method : str; {'ar1asym','ar1sim'}

            Method to generate surrogates. AR1sim uses simulated timeseries with similar persistence. AR1asymp represents the closed form solution. The default is AR1sim

        seed : int, optional

            Option to set the seed for reproducibility. The default is None.

        qs : list, optional

            Significance levels to return. The default is [0.95].

        settings : dict, optional

            Parameters for the specific significance test. The default is None. Note that the default value for the asymptotic solution is `time-average`

        scalogram : pyleoclim.Scalogram object, optional

            Scalogram containing signif_scals exported during significance testing of scalogram.
            If number is None and signif_scals are present, will use length of scalogram list as number of significance tests

        Returns
        -------

        new : pyleoclim.core.psds.PSD

            New PSD object with appropriate significance test

        Examples
        --------
        
        Compute the spectrum of the Southern Oscillation Index and assess significance against an AR(1) benchmark:

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pandas as pd
            csv = pd.read_csv('https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/master/example_data/soi_data.csv',skiprows = 1)
            soi = pyleo.Series(time = csv['Year'],value = csv['Value'], time_name = 'Years', time_unit = 'AD')
            psd = soi.standardize().spectral('mtm',settings={'NW':2})
            psd_sim = psd.signif_test(number=20)
            @savefig psd_sim.png
            fig, ax = psd_sim.plot()
            pyleo.closefig(fig)
            
        By default, this method uses 200 Monte Carlo simulations of an AR(1) process. 
        For a smoother benchmark, up the number of simulations. 
        Also, you may obtain and visualize several quantiles at once, e.g. 90% and 95%:
        
        .. ipython:: python
            :okwarning:
            :okexcept:
                
            psd_1000 = psd.signif_test(number=100, qs=[0.90, 0.95])
            @savefig psd_1000.png
            fig, ax = psd_1000.plot()
            pyleo.closefig(fig)
        
        Another option is to use a closed-form, asymptotic solution for the AR(1) spectrum:
            
        .. ipython:: python
            :okwarning:
            :okexcept:
            
            psd_asym = psd.signif_test(method='ar1asym',qs=[0.90, 0.95])
            @savefig psd_asym.png
            fig, ax = psd_asym.plot()
            pyleo.closefig(fig)
            
        If significance tests from a comparable scalogram have been saved, they can be passed here to speed up the generation of noise realizations for significance testing.
        Setting export_scal to True saves the noise realizations generated during significance testing for future use:
            
        .. ipython:: python
            :okwarning:
            :okexcept:
                
            scalogram = soi.standardize().wavelet().signif_test(number=20, export_scal=True)

        The psd can be calculated by using the previously generated scalogram
        
        .. ipython:: python
            :okwarning:
            :okexcept:
                
            psd_scal = soi.standardize().spectral(scalogram=scalogram)

        The same scalogram can then be passed to do significance testing. 
        Pyleoclim will dig through the scalogram object to find the saved 
        noise realizations and reuse them flexibly.
        
        .. ipython:: python
            :okwarning:
            :okexcept:
                
            @savefig psd_scal.png
            fig, ax = psd.signif_test(scalogram=scalogram).plot()
            pyleo.closefig(fig)

        See also
        --------

        pyleoclim.utils.wavelet.tc_wave_signif : asymptotic significance calculation

        pyleoclim.core.psds.MultiplePSD : Object storing several PSDs from different Series or ensemble members in an age model

        pyleoclim.core.scalograms.Scalogram :  Scalogram object

        pyleoclim.core.series.Series.surrogates : Generate surrogates with increasing time axis

        pyleoclim.core.series.Series.spectral : Performs spectral analysis on Pyleoclim Series

        pyleoclim.core.series.Series.wavelet : Performs wavelet analysis on Pyleoclim Series

        '''

        if self.spec_method == 'wwz' and method == 'ar1asym':
            raise ValueError('Asymptotic solution is not supported for the wwz method')

        if self.spec_method == 'lomb_scargle' and method == 'ar1asym':
            raise ValueError('Asymptotic solution is not supported for the Lomb-Scargle method')

        if method not in ['ar1sim', 'ar1asym']:
                raise ValueError("The available methods are 'ar1sim' and 'ar1asym'")

        if method == 'ar1sim':
            signif_scals = None
            if scalogram:
                try:
                    signif_scals = scalogram.signif_scals
                except:
                    return ValueError('Could not find signif_scals in passed object, make sure this is a scalogram with signif_scals that were saved during significance testing')


            if number is None and signif_scals:
                number = len(signif_scals.scalogram_list)
            elif number is None and signif_scals is None:
                number = 200
            elif number == 0:
                return self

            new = self.copy()
            surr = self.timeseries.surrogates(
                number=number, seed=seed, method=method, settings=settings
            )

            if signif_scals:
                surr_psd = surr.spectral(
                    method=self.spec_method, settings=self.spec_args, scalogram_list=signif_scals
                )
            else:
                surr_psd = surr.spectral(method=self.spec_method, settings=self.spec_args)
            new.signif_qs = surr_psd.quantiles(qs=qs)
            new.signif_method = method

        elif method == 'ar1asym':
            std = self.timeseries.stats()['std'] # assess standard deviation 
            if np.abs(std-1) > 0.1: 
                warnings.warn("Asymptoics are only defined for a standard deviation of unity. Please apply to a standardized series only")
            
            new=self.copy()

            if type(qs) is not list:
                raise TypeError('qs should be a list')

            settings = {'sigtest':'time-average'} if settings is None else settings.copy()

            if self.spec_method=='cwt':
                if 'dof' not in settings.keys():
                    dof = len(self.timeseries.value) - self.spec_args['scale']
                    settings.update({'dof':dof})
                signif_levels=waveutils.tc_wave_signif(self.timeseries.value,
                                                       self.timeseries.time,
                                                       self.spec_args['scale'],
                                                       self.spec_args['mother'],
                                                       self.spec_args['param'],
                                                       qs=qs, **settings)
            else:
                # hard code Mortlet values to obtain the spectrum
                param = 6
                fourier_factor = 4 * np.pi / (param + np.sqrt(2 + param**2))
                scale = 1/(fourier_factor*self.frequency)
                if 'dof' not in settings.keys():
                    dof = len(self.timeseries.value) - scale
                    settings.update({'dof':dof})
                signif_levels=waveutils.tc_wave_signif(self.timeseries.value,
                                                       self.timeseries.time,
                                                       scale,
                                                       'MORLET',
                                                       param,
                                                       qs=qs, **settings)

            # get it back into the object
            new.signif_method = method

            ms_base = []
            for idx, item in enumerate(signif_levels):
                label = str(int(qs[idx]*100))+'%'
                s = PSD(frequency=self.frequency, amplitude = item, label=label)
                ms_base.append(s)
                new.signif_qs = MultiplePSD(ms_base)

        return new

    def beta_est(self, fmin=None, fmax=None, logf_binning_step='max', verbose=False):
        ''' Estimate the scaling exponent (beta) of the PSD 

        For a power law S(f) ~ f^beta in log-log space, beta is simply the slope.
        
        Parameters
        ----------

        fmin : float, optional
            the minimum frequency edge for beta estimation; the default is the minimum of the frequency vector of the PSD obj

        fmax : float, optional
            the maximum frequency edge for beta estimation; the default is the maximum of the frequency vector of the PSD obj

        logf_binning_step : str, {'max', 'first'}
            if 'max', then the maximum spacing of log(f) will be used as the binning step
            if 'first', then the 1st spacing of log(f) will be used as the binning step

        verbose : bool; {True, False}
            If True, will print warning messages if there is any

        Returns
        -------

        new : pyleoclim.core.psds.PSD
            New PSD object with the estimated scaling slope information, which is stored as a dictionary that includes:
            - beta: the scaling factor
            - std_err: the one standard deviation error of the scaling factor
            - f_binned: the binned frequency series, used as X for linear regression
            - psd_binned: the binned PSD series, used as Y for linear regression
            - Y_reg: the predicted Y from linear regression, used with f_binned for the slope curve plotting

        Examples
        --------

        Generate fractal noise and verify that its scaling exponent is close to unity

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            t, v = pyleo.utils.tsmodel.gen_ts(model='colored_noise')
            ts = pyleo.Series(time=t, value= v, label = 'fractal noise, unit slope')
            psd = ts.detrend().spectral(method='cwt')

            # estimate the scaling slope
            psd_beta = psd.beta_est(fmin=1/50, fmax=1/2)

            @savefig color_noise_beta.png
            fig, ax = psd_beta.plot(color='tab:blue',beta_kwargs={'color':'tab:red','linewidth':2})
            pyleo.closefig(fig)

        See also
        --------
        pyleoclim.core.series.Series.spectral : spectral analysis

        pyleoclim.utils.spectral.beta_estimation : Estimate the scaling exponent of a power spectral density
        
        pyleoclim.core.psds.PSD.plot : plotting method for PSD objects

        '''
        if fmin is None:
            fmin = np.min(self.frequency)

        if fmax is None:
            fmax = np.max(self.frequency)

        res = specutils.beta_estimation(self.amplitude, self.frequency, fmin=fmin, fmax=fmax, logf_binning_step=logf_binning_step, verbose=verbose)
        res_dict = {
            'beta': res.beta,
            'std_err': res.std_err,
            'f_binned': res.f_binned,
            'psd_binned': res.psd_binned,
            'Y_reg': res.Y_reg,
        }
        new = self.copy()
        new.beta_est_res = res_dict
        return new

    def anti_alias(self, avgs=2):
        ''' Apply the anti-aliasing filter

        Parameters
        ----------

        avgs : int
            flag for whether spectrum is derived from instantaneous point measurements (avgs<>1)
            OR from measurements averaged over each sampling interval (avgs==1)

        Returns
        -------

        new : pyleoclim.core.psds.PSD
            New PSD object with the spectral aliasing effect alleviated.

        Examples
        --------

        Generate colored noise with scaling exponent equals to unity, and test the impact of anti-aliasing filter

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo

            t, v = pyleo.utils.tsmodel.gen_ts('colored_noise', alpha=1, m=1e5) # m=1e5 leads to aliasing
            ts = pyleo.Series(time=t, value=v, label='colored noise')

            # without the anti-aliasing filter
            @savefig color_noise_no_anti_alias.png
            fig, ax = ts.spectral(method='mtm').beta_est().plot()

            # with the anti-aliasing filter
            @savefig color_noise_anti_alias.png
            fig, ax = ts.spectral(method='mtm').anti_alias().beta_est().plot()

        References
        ----------

        Kirchner, J. W. Aliasing in 1/f(alpha) noise spectra: origins, consequences, and remedies.
        Phys Rev E Stat Nonlin Soft Matter Phys 71, 66110 (2005).

        See also
        --------

        pyleoclim.utils.wavelet.AliasFilter.alias_filter : anti-aliasing filter
        '''
        new = self.copy()

        dt = np.median(np.diff(self.timeseries.time))
        f_sampling = 1/dt
        psd_copy = self.amplitude[1:]
        freq_copy = self.frequency[1:]
        alpha, filtered_pwr, model_pwer, aliased_pwr = waveutils.AliasFilter().alias_filter(
            freq_copy, psd_copy, f_sampling, f_sampling*1e3, np.min(self.frequency), avgs)

        new.amplitude[1:] = np.copy(filtered_pwr)
        return new

    def plot(self, in_loglog=True, in_period=True, label=None, xlabel=None, ylabel='PSD', title=None,
             marker=None, markersize=None, color=None, linestyle=None, linewidth=None, transpose=False,
             xlim=None, ylim=None, figsize=[10, 4], savefig_settings=None, ax=None,
             legend=True, lgd_kwargs=None, xticks=None, yticks=None, alpha=None, zorder=None,
             plot_kwargs=None, signif_clr='red', signif_linestyles=['--', '-.', ':'], signif_linewidth=1,
             plot_beta=True, beta_kwargs=None):

        '''Plots the PSD estimates and signif level if included


        Parameters
        ----------
        
        in_loglog : bool; {True, False}, optional

            Plot on loglog axis. The default is True.

        in_period : bool; {True, False}, optional

            Plot the x-axis as periodicity rather than frequency. The default is True.

        label : str, optional

            label for the series. The default is None.

        xlabel : str, optional

            Label for the x-axis. The default is None. Will guess based on Series

        ylabel : str, optional

            Label for the y-axis. The default is 'PSD'.

        title : str, optional

            Plot title. The default is None.

        marker : str, optional

            marker to use. The default is None.

        markersize : int, optional

            size of the marker. The default is None.

        color : str, optional

            Line color. The default is None.

        linestyle : str, optional

            linestyle. The default is None.

        linewidth : float, optional

            Width of the line. The default is None.

        transpose : bool; {True, False}, optional

            Plot periodicity on y-. The default is False.

        xlim : list, optional

            x-axis limits. The default is None.

        ylim : list, optional

            y-axis limits. The default is None.

        figsize : list, optional

            Figure size. The default is [10, 4].

        savefig_settings : dict, optional

            save settings options. The default is None.
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existed or non-existed path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}

        ax : ax, optional

            The matplotlib.Axes object onto which to return the plot.
            The default is None.

        legend : bool; {True, False}, optional

            whether to plot the legend. The default is True.

        lgd_kwargs : dict, optional

            Arguments for the legend. The default is None.

        xticks : list, optional

            xticks to use. The default is None.

        yticks : list, optional

            yticks to use. The default is None.

        alpha : float, optional

            Transparency setting. The default is None.

        zorder : int, optional

            Order for the plot. The default is None.

        plot_kwargs : dict, optional

            Other plotting argument. The default is None.

        signif_clr : str, optional

            Color for the significance line. The default is 'red'.

        signif_linestyles : list of str, optional

            Linestyles for significance. The default is ['--', '-.', ':'].

        signif_linewidth : float, optional

            width of the significance line. The default is 1.

        plot_beta : bool; {True, False}, optional

            If True and self.beta_est_res is not None, then the scaling slope line will be plotted

        beta_kwargs : dict, optional

            The visualization keyword arguments for the scaling slope

        Returns
        -------
        
        fig, ax

        Examples
        --------

        Generate fractal noise, assess significance against an AR(1) benchmark, and plot:

        .. ipython:: python
            :okwarning:
            :okexcept:

            import matplotlib.pyplot as plt
            
            t, v = pyleo.utils.tsmodel.gen_ts(model='colored_noise')
            ts = pyleo.Series(time = t, value = v, label = 'fractal noise')
            tsn = ts.standardize()
    
            psd_sim = tsn.spectral(method='mtm').signif_test(number=20)
            @savefig mtm_sim.png
            psd_sim.plot()
            pyleo.closefig(fig)
            
        If you add the estimate of the scaling exponent, the line of best fit 
        will be added to the plot, and the estimated exponent to its legend. For instance:
            
        .. ipython:: python
            :okwarning:
            :okexcept:    
            
            psd_beta = psd_sim.beta_est(fmin=1/100, fmax=1/2)

            @savefig mtm_sig_beta.png
            fig, ax = psd_beta.plot()
            pyleo.closefig(fig)

        See also
        --------

        pyleoclim.core.series.Series.spectral : spectral analysis
        
        pyleoclim.core.psds.PSD.signif_test : significance testing for PSD objects
        
        pyleoclim.core.psds.PSD.beta_est : scaling exponent estimation for PSD objects

        '''
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        plot_kwargs = self.plot_kwargs if plot_kwargs is None else plot_kwargs.copy()
        beta_kwargs = {} if beta_kwargs is None else beta_kwargs.copy()
        lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

        if label is None:
            if plot_beta and self.beta_est_res is not None:
                label = fr'{self.label} ($\hat \beta=${self.beta_est_res["beta"]:.2f}$\pm${self.beta_est_res["std_err"]:.2f})'
            else:
                label = self.label

        if label is not None:
            plot_kwargs.update({'label': label})

        if marker is not None:
            plot_kwargs.update({'marker': marker})

        if markersize is not None:
            plot_kwargs.update({'markersize': markersize})

        if color is not None:
            plot_kwargs.update({'color': color})

        if linestyle is not None:
            plot_kwargs.update({'linestyle': linestyle})

        if linewidth is not None:
            plot_kwargs.update({'linewidth': linewidth})

        if alpha is not None:
            plot_kwargs.update({'alpha': alpha})

        if zorder is not None:
            plot_kwargs.update({'zorder': zorder})

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if in_period:
            idx = np.argwhere(self.frequency==0)
            x_axis = 1/np.delete(self.frequency, idx)
            y_axis = np.delete(self.amplitude, idx)
            if xlabel is None:
                xlabel = f'Period [{self.period_unit}]' if self.period_unit is not None else 'Period'

            if xticks is None:
                xticks_default = np.array([0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 1e4, 2e4, 5e4, 1e5, 2e5, 5e5, 1e6])
                mask = (xticks_default >= np.nanmin(x_axis)) & (xticks_default <= np.nanmax(x_axis))
                xticks = xticks_default[mask]

            if xlim is None:
                xlim = [np.max(xticks), np.min(xticks)]

        else:
            idx = np.argwhere(self.frequency==0)
            x_axis = np.delete(self.frequency, idx)
            y_axis = np.delete(self.amplitude, idx)
            if xlabel is None:
                xlabel = f'Frequency [1/{self.period_unit}]' if self.period_unit is not None else 'Frequency'

            if xlim is None:
                xlim = ax.get_xlim()
                xlim = [np.min(xlim), np.max(xlim)]

        if transpose:
            ylim = [np.min(x_axis), np.max(x_axis)]
            xlim = [np.min(y_axis), np.max(y_axis)]
            x_axis, y_axis = y_axis, x_axis
            ax.plot(x_axis,y_axis, **plot_kwargs)

            xticks, yticks = yticks, xticks
            xlabel, ylabel = ylabel, xlabel
        else:
            ax.set_xlim(xlim)
            ax.plot(x_axis, y_axis, **plot_kwargs)

        #
        # if transpose:
        #     x_axis, y_axis = y_axis, x_axis
        #     xlim, ylim = ylim, xlim
        #     xticks, yticks = yticks, xticks
        #     xlabel, ylabel = ylabel, xlabel
        #     ax.set_ylim(ylim[::-1])
        #     # print('thing',ylim[::-1])
        # else:
        #     ax.set_xlim(xlim)
        # ax.plot(x_axis, y_axis, **plot_kwargs)


        # plot significance levels
        if self.signif_qs is not None:
            signif_method_label = {
                'ar1sim': 'AR(1) simulations',
                'ar1asym': 'AR(1) asymptotic solution'
            }
            nqs = np.size(self.signif_qs.psd_list)

            for i, q in enumerate(self.signif_qs.psd_list):
                idx = np.argwhere(q.frequency==0)
                signif_x_axis = 1/np.delete(q.frequency, idx) if in_period else np.delete(q.frequency, idx)
                signif_y_axis = np.delete(q.amplitude, idx)
                if transpose:
                    signif_x_axis, signif_y_axis = signif_y_axis, signif_x_axis

                ax.plot(
                    signif_x_axis, signif_y_axis,
                    label=f'{signif_method_label[self.signif_method]}, {q.label} threshold',
                    color=signif_clr,
                    linestyle=signif_linestyles[i%3],
                    linewidth=signif_linewidth,
                )

        if in_loglog:
            ax.set_xscale('log')
            ax.set_yscale('log')

        if xticks is not None:
            ax.set_xticks(xticks)
            ax.xaxis.set_major_formatter(ScalarFormatter())
            ax.xaxis.set_major_formatter(FormatStrFormatter('%g'))

        if yticks is not None:
            ax.set_yticks(yticks)
            ax.yaxis.set_major_formatter(ScalarFormatter())
            ax.yaxis.set_major_formatter(FormatStrFormatter('%g'))

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        if plot_beta and self.beta_est_res is not None:
            plot_beta_kwargs = {
                'linestyle': '--',
                'color': 'k',
                'linewidth': 1,
                'zorder': 99,
            }
            plot_beta_kwargs.update(beta_kwargs)
            beta_x_axis = 1/self.beta_est_res['f_binned']
            beta_y_axis = self.beta_est_res['Y_reg']
            if transpose:
                beta_x_axis, beta_y_axis = beta_y_axis, beta_x_axis
            ax.plot(beta_x_axis, beta_y_axis , **plot_beta_kwargs)

        if legend:
            lgd_args = {'frameon': False}
            lgd_args.update(lgd_kwargs)
            ax.legend(**lgd_args)

        if title is not None:
            ax.set_title(title)

        if xlim is not None:
            ax.set_xlim(xlim)

        if ylim is not None:
            ax.set_ylim(ylim)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax
        

import seaborn as sns
import matplotlib as mpl
from scipy.stats.mstats import mquantiles

class MultiplePSD:
    '''MultiplePSD objects store several PSDs from different Series or ensemble members from 
    a posterior distribution (e.g. age model, Bayesian climate reconstruction, etc). 
    This is used extensively for Monte Carlo significance tests. 
    '''
    def __init__(self, psd_list, beta_est_res=None):
        ''' Object for multiple PSD.

        This object stores several PSDs from different Series or ensemble members in an age model.         
       
        Parameters
        ----------
        
        beta_est_res : numpy.array
        
            Results of the beta estimation calculation
        
        See also
        --------
        
        pyleoclim.core.psds.PSD.beta_est : Calculates the scaling exponent (i.e., the slope in a log-log plot) of the spectrum (beta)

        '''
        self.psd_list = psd_list
        if beta_est_res is None:
            self.beta_est_res = beta_est_res
        else:
            self.beta_est_res = np.array(beta_est_res)

    def copy(self):
        '''Copy object
        '''
        return deepcopy(self)

    def quantiles(self, qs=[0.05, 0.5, 0.95], lw=[0.5, 1.5, 0.5]):
        
        '''Calculate the quantiles of the significance testing

        Parameters
        ----------
        
        qs : list, optional
        
            List of quantiles to consider for the calculation. The default is [0.05, 0.5, 0.95].
            
        lw : list, optional
        
            Linewidth to use for plotting each level. Should be the same length as qs. The default is [0.5, 1.5, 0.5].

        Raises
        ------
        
        ValueError
        
            Frequency axis not consistent across the PSD list!

        Returns
        -------
        
        psds : pyleoclim.core.psds.MultiplePSD

        '''
        if self.psd_list[0].timeseries is not None:
            period_unit = self.psd_list[0].timeseries.time_unit

        freq = np.copy(self.psd_list[0].frequency)
        amps = []
        for psd in self.psd_list:
            if not np.array_equal(psd.frequency, freq):
                raise ValueError('Frequency axis not consistent across the PSD list!')

            amps.append(psd.amplitude)

        amps = np.array(amps)
        amp_qs = mquantiles(amps, qs, axis=0)

        psd_list = []
        for i, amp in enumerate(amp_qs):
            psd_tmp = PSD(frequency=freq, amplitude=amp, label=f'{qs[i]*100:g}%', plot_kwargs={'color': 'gray', 'linewidth': lw[i]}, period_unit=period_unit)
            psd_list.append(psd_tmp)

        psds = MultiplePSD(psd_list=psd_list)
        return psds

    def beta_est(self, fmin=None, fmax=None, logf_binning_step='max', verbose=False):
        
        ''' Estimate the scaling exponent of each constituent PSD 
        
        This function calculates the scaling exponent (beta) for each of the PSDs stored in the object. 
        The scaling exponent represents the slope of the spectrum in log-log space. 

        Parameters
        ----------

        fmin : float
        
            the minimum frequency edge for beta estimation; the default is the minimum of the frequency vector of the PSD object

        fmax : float
        
            the maximum frequency edge for beta estimation; the default is the maximum of the frequency vector of the PSD object

        logf_binning_step : str; {'max', 'first'}
        
            if 'max', then the maximum spacing of log(f) will be used as the binning step.
            if 'first', then the 1st spacing of log(f) will be used as the binning step.

        verbose : bool
        
            If True, will print warning messages if there is any

        Returns
        -------

        new : pyleoclim.core.psds.MultiplePSD
        
            New MultiplePSD object with the estimated scaling slope information, which is stored as a dictionary that includes:
            - beta: the scaling factor
            - std_err: the one standard deviation error of the scaling factor
            - f_binned: the binned frequency series, used as X for linear regression
            - psd_binned: the binned PSD series, used as Y for linear regression
            - Y_reg: the predicted Y from linear regression, used with f_binned for the slope curve plotting

        See also
        --------

        pyleoclim.core.psds.PSD.beta_est : scaling exponent estimation for a single PSD object

        '''

        res_dict = {}
        res_dict['beta'] = []
        res_dict['std_err'] = []
        res_dict['f_binned'] = []
        res_dict['psd_binned'] = []
        res_dict['Y_reg'] = []
        psd_beta_list = []
        for psd_obj in self.psd_list:
            psd_beta = psd_obj.beta_est(fmin=fmin, fmax=fmax, logf_binning_step=logf_binning_step, verbose=verbose)
            psd_beta_list.append(psd_beta)
            res = psd_beta.beta_est_res
            for k in res_dict.keys():
                res_dict[k].append(res[k])

        new = self.copy()
        new.beta_est_res = res_dict
        new.psd_list = psd_beta_list
        return new


    def plot(self, figsize=[10, 4], in_loglog=True, in_period=True, xlabel=None, ylabel='PSD', title=None,
             xlim=None, ylim=None, savefig_settings=None, ax=None, xticks=None, yticks=None, legend=True,
             colors=None, cmap=None, norm=None, plot_kwargs=None, lgd_kwargs=None):
        '''Plot multiple PSDs on the same plot

        Parameters
        ----------
        
        figsize : list, optional
        
            Figure size. The default is [10, 4].
            
        in_loglog : bool, optional
        
            Whether to plot in loglog. The default is True.
            
        in_period : bool, {True, False} optional
        
            Plots against periods instead of frequencies. The default is True.
            
        xlabel : str, optional
        
            x-axis label. The default is None.
            
        ylabel : str, optional
        
            y-axis label. The default is 'Amplitude'.
            
        title : str, optional
        
            Title for the figure. The default is None.
            
        xlim : list, optional
        
            Limits for the x-axis. The default is None.
            
        ylim : list, optional
        
            limits for the y-axis. The default is None.
            
        colors : a list of, or one, Python supported color code (a string of hex code or a tuple of rgba values)
        
            Colors for plotting.
            If None, the plotting will cycle the 'tab10' colormap;
            if only one color is specified, then all curves will be plotted with that single color;
            if a list of colors are specified, then the plotting will cycle that color list.
            
        cmap : str
        
            The colormap to use when "colors" is None.
            
        norm : matplotlib.colors.Normalize like
        
            The nomorlization for the colormap.
            If None, a linear normalization will be used.
            
        savefig_settings : dict, optional
        
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existing or non-existing path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}
            
        ax : matplotlib axis, optional
        
            The matplotlib axis object on which to retrun the figure. The default is None.
            
        xticks : list, optional
        
            x-ticks label. The default is None.
            
        yticks : list, optional
        
            y-ticks label. The default is None.
            
        legend : bool, optional
        
            Whether to plot the legend. The default is True.
            
        plot_kwargs : dictionary, optional
        
            Parameters for plot function. The default is None.
            
        lgd_kwargs : dictionary, optional
        
            Parameters for legend. The default is None.

        Returns
        -------
        
        fig : matplotlib.pyplot.figure
        
        ax : matplotlib.pyplot.axis
        
        See also
        --------
        
        pyleoclim.core.psds.PSD.plot : plotting method for PSD objects

        '''
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
        lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        for idx, psd in enumerate(self.psd_list):

            tmp_plot_kwargs = {}
            if psd.plot_kwargs is not None:
                tmp_plot_kwargs.update(psd.plot_kwargs)

            tmp_plot_kwargs.update(plot_kwargs)

            # get color for each psd curve
            use_clr = False

            if 'color' not in tmp_plot_kwargs and 'c' not in 'tmp_plot_kwargs':
                use_clr = True

            if 'color' in tmp_plot_kwargs and tmp_plot_kwargs['color'] is None:
                use_clr = True

            if 'c' in tmp_plot_kwargs and tmp_plot_kwargs['c'] is None:
                use_clr = True

            if colors is not None or cmap is not None:
                use_clr = True

            if use_clr:
                # use the color based on the argument 'colors' or 'cmap'
                if colors is None:
                    cmap = 'tab10' if cmap is None else cmap
                    cmap_obj = plt.get_cmap(cmap)
                    if hasattr(cmap_obj, 'colors'):
                        nc = len(cmap_obj.colors)
                    else:
                        nc = len(self.psd_list)

                    if norm is None:
                        norm = mpl.colors.Normalize(vmin=0, vmax=nc-1)

                    clr = cmap_obj(norm(idx%nc))
                elif type(colors) is str:
                    clr = colors
                elif type(colors) is list:
                    nc = len(colors)
                    clr = colors[idx%nc]
                else:
                    raise TypeError('"colors" should be a list of, or one, Python supported color code (a string of hex code or a tuple of rgba values)')

                tmp_plot_kwargs.update({'color': clr})

            ax = psd.plot(
                figsize=figsize, in_loglog=in_loglog, in_period=in_period, xlabel=xlabel, ylabel=ylabel,
                title=title, xlim=xlim, ylim=ylim, savefig_settings=savefig_settings, ax=ax,
                xticks=xticks, yticks=yticks, legend=legend, plot_kwargs=tmp_plot_kwargs, lgd_kwargs=lgd_kwargs,
            )

        if title is not None:
            ax.set_title(title)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax

    def anti_alias(self, avgs=2, mute_pbar=False):
        ''' Apply the anti-aliasing filter

        Parameters
        ----------

        avgs : int
        
            flag for whether spectrum is derived from instantaneous point measurements (avgs<>1)
            OR from measurements averaged over each sampling interval (avgs==1)

        mute_pbar : bool; {True,False}
        
            If True, the progressbar will be muted. Default is False.

        Returns
        -------

        new : pyleoclim.core.psds.MultiplePSD
            New MultiplePSD object with the spectral aliasing effect alleviated.


        References
        ----------

        Kirchner, J. W. Aliasing in 1/f(alpha) noise spectra: origins, consequences, and remedies.
        Phys Rev E Stat Nonlin Soft Matter Phys 71, 66110 (2005).

        See also
        --------

        pyleoclim.utils.wavelet.AliasFilter.alias_filter : anti-aliasing filter
        '''
        psd_aa_list = []
        for psd_obj in tqdm(self.psd_list,  total=len(self.psd_list), disable=mute_pbar, desc='Applying the anti-alias filter'):
            psd_aa = psd_obj.anti_alias(avgs=avgs)
            psd_aa_list.append(psd_aa)

        new = self.copy()
        new.psd_list = psd_aa_list
        return new



    def plot_envelope(self, figsize=[10, 4], qs=[0.025, 0.5, 0.975],
             in_loglog=True, in_period=True, xlabel=None, ylabel='PSD', title=None,
             xlim=None, ylim=None, savefig_settings=None, ax=None, xticks=None, yticks=None, plot_legend=True,
             curve_clr=sns.xkcd_rgb['pale red'], curve_lw=3, shade_clr=sns.xkcd_rgb['pale red'], shade_alpha=0.3, shade_label=None,
             lgd_kwargs=None, members_plot_num=10, members_alpha=0.3, members_lw=1, seed=None):

        '''Plot an envelope statistics for mulitple PSD
        
        This function plots an envelope statistics from multiple PSD. This is especially useful when the PSD are coming from an ensemble of possible solutions (e.g., age ensembles)

        Parameters
        ----------
        
        figsize : list, optional
        
            The figure size. The default is [10, 4].
            
        qs : list, optional
        
            The significance levels to consider. The default is [0.025, 0.5, 0.975].
            
        in_loglog : bool, optional
        
            Plot in log space. The default is True.
            
        in_period : bool, optional
        
            Whether to plot periodicity instead of frequency. The default is True.
            
        xlabel : str, optional
        
            x-axis label. The default is None.
            
        ylabel : str, optional
        
            y-axis label. The default is 'Amplitude'.
            
        title : str, optional
        
            Plot title. The default is None.
            
        xlim : list, optional
        
            x-axis limits. The default is None.
            
        ylim : list, optional
        
            y-axis limits. The default is None.
            
        savefig_settings : dict, optional
        
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existing or non-existing path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"} The default is None.
            
        ax : matplotlib.ax, optional
        
            Matplotlib axis on which to return the plot. The default is None.
            
        xticks : list, optional
        
            xticks label. The default is None.
            
        yticks : list, optional
        
            yticks label. The default is None.
            
        plot_legend : bool, optional
        
            Wether to plot the legend. The default is True.
            
        curve_clr : str, optional
        
            Color of the main PSD. The default is sns.xkcd_rgb['pale red'].
            
        curve_lw : str, optional
        
            Width of the main PSD line. The default is 3.
            
        shade_clr : str, optional
        
            Color of the shaded envelope. The default is sns.xkcd_rgb['pale red'].
            
        shade_alpha : float, optional
        
            Transparency on the envelope. The default is 0.3.
            
        shade_label : str, optional
        
            Label for the envelope. The default is None.
            
        lgd_kwargs : dict, optional
        
            Parameters for the legend. The default is None.
            
        members_plot_num : int, optional
        
            Number of individual members to plot. The default is 10.
            
        members_alpha : float, optional
        
            Transparency of the lines representing the multiple members. The default is 0.3.
            
        members_lw : float, optional
        
            With of the lines representing the multiple members. The default is 1.
            
        seed : int, optional
        
            Set the seed for random number generator. Useful for reproducibility. The default is None.

        Returns
        -------
        fig : matplotlib.pyplot.figure
        
        ax : matplotlib.pyplot.axis
        
        See also
        --------
        
        pyleoclim.core.psds.PSD.plot : plotting method for PSD objects
        
        pyleoclim.core.ensembleseries.EnsembleSeries.plot_envelope : envelope plot for ensembles
        
        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:
            
            import pyleoclim as pyleo
            import numpy as np
            nn = 30 # number of noise realizations
            nt = 500 # timeseries length
            psds = []

            time, signal = pyleo.utils.gen_ts(model='colored_noise',nt=nt,alpha=1.0)
            
            ts = pyleo.Series(time=time, value = signal).standardize()
            noise = np.random.randn(nt,nn)

            for idx in range(nn):  # noise
                ts = pyleo.Series(time=time, value=signal+10*noise[:,idx])
                psd = ts.spectral()
                psds.append(psd)

            mPSD = pyleo.MultiplePSD(psds)
            
            @savefig ens_specplot.png
            fig, ax = mPSD.plot_envelope()
            pyleo.closefig(fig) 

        '''
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if members_plot_num > 0:
            if seed is not None:
                np.random.seed(seed)

            npsd = np.size(self.psd_list)
            random_draw_idx = np.random.choice(npsd, members_plot_num)

            for idx in random_draw_idx:
                self.psd_list[idx].plot(
                    in_loglog=in_loglog, in_period=in_period, xlabel=xlabel, ylabel=ylabel,
                    xlim=xlim, ylim=ylim, xticks=xticks, yticks=yticks, ax=ax, color='gray', alpha=members_alpha,
                    zorder=99, linewidth=members_lw,
                )
            ax.plot(np.nan, np.nan, color='gray', label=f'example members (n={members_plot_num})')

        psd_qs = self.quantiles(qs=qs)
        psd_qs.psd_list[1].plot(
            in_loglog=in_loglog, in_period=in_period, xlabel=xlabel, ylabel=ylabel, linewidth=curve_lw,
            xlim=xlim, ylim=ylim, xticks=xticks, yticks=yticks, ax=ax, color=curve_clr, zorder=100
        )


        if in_period:
            x_axis = 1/psd_qs.psd_list[0].frequency
        else:
            x_axis = psd_qs.psd_list[0].frequency

        if shade_label is None:
            shade_label = f'{psd_qs.psd_list[0].label}-{psd_qs.psd_list[-1].label}'

        ax.fill_between(
            x_axis, psd_qs.psd_list[0].amplitude, psd_qs.psd_list[-1].amplitude,
            color=shade_clr, alpha=shade_alpha, edgecolor=shade_clr, label=shade_label,
        )

        if title is not None:
            ax.set_title(title)

        if plot_legend:
            lgd_args = {'frameon': False}
            lgd_args.update(lgd_kwargs)
            ax.legend(**lgd_args)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax
    



''' Tests for pyleoclim.core.ui.EnsembleSeries

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''
import numpy as np
import pandas as pd

from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

import pytest

import pyleoclim as pyleo
from pyleoclim.utils.tsmodel import (
    ar1_sim,
    colored_noise,
)
from pyleoclim.tests.examples import load_dataset

# a collection of useful functions

def gen_ts(model, nt, alpha):
    'wrapper for gen_ts in pyleoclim'

    t, v = pyleo.utils.gen_ts(model=model, nt=nt, alpha=alpha)
    ts = pyleo.Series(t, v)
    return ts


def gen_normal(loc=0, scale=1, nt=100):
    ''' Generate random data with a Gaussian distribution
    '''
    t = np.arange(nt)
    v = np.random.normal(loc=loc, scale=scale, size=nt)
    return t, v

def gen_colored_noise(alpha=1, nt=100, f0=None, m=None, seed=None):
    ''' Generate colored noise
    '''
    t = np.arange(nt)
    v = colored_noise(alpha=alpha, t=t, f0=f0, m=m, seed=seed)
    return t, v


# Tests below
class TestUIEnsembleSeriesCorrelation():
    def test_correlation_t0(self):
        '''Test for EnsembleSeries.correlation() when the target is a Series
        '''
        nt = 100
        t0, v0 = gen_colored_noise(nt=nt)
        t0, noise = gen_normal(nt=nt)

        ts0 = pyleo.Series(time=t0, value=v0)
        ts1 = pyleo.Series(time=t0, value=v0+noise)
        ts2 = pyleo.Series(time=t0, value=v0+2*noise)

        ts_list = [ts1, ts2]

        ts_ens = pyleo.EnsembleSeries(ts_list)

        corr_res = ts_ens.correlation(ts0)
        signif_list = corr_res.signif
        for signif in signif_list:
            assert signif is True


    def test_correlation_t1(self):
        '''Test for EnsembleSeries.correlation() when the target is an EnsembleSeries with same number of Series
        '''
        nt = 100
        t0, v0 = gen_colored_noise(nt=nt)
        t0, noise = gen_normal(nt=nt)

        ts0 = pyleo.Series(time=t0, value=v0)
        ts1 = pyleo.Series(time=t0, value=v0+noise)
        ts2 = pyleo.Series(time=t0, value=v0+2*noise)
        ts3 = pyleo.Series(time=t0, value=v0+1/2*noise)

        ts_list1 = [ts0, ts1]
        ts_list2 = [ts2, ts3]

        ts_ens = pyleo.EnsembleSeries(ts_list1)
        ts_target = pyleo.EnsembleSeries(ts_list2)

        corr_res = ts_ens.correlation(ts_target)
        signif_list = corr_res.signif
        for signif in signif_list:
            assert signif is True

        assert np.size(corr_res.p) == np.size(ts_list1)


    def test_correlation_t2(self):
        '''Test for EnsembleSeries.correlation() when the target is an EnsembleSeries with fewer Series
        '''
        nt = 100
        t0, v0 = gen_colored_noise(nt=nt)
        t0, noise = gen_normal(nt=nt)

        ts0 = pyleo.Series(time=t0, value=v0)
        ts1 = pyleo.Series(time=t0, value=v0+noise)
        ts2 = pyleo.Series(time=t0, value=v0+2*noise)
        ts3 = pyleo.Series(time=t0, value=v0+1/2*noise)
        ts4 = pyleo.Series(time=t0, value=v0+3/2*noise)

        ts_list1 = [ts0, ts1, ts4]
        ts_list2 = [ts2, ts3]

        ts_ens = pyleo.EnsembleSeries(ts_list1)
        ts_target = pyleo.EnsembleSeries(ts_list2)

        corr_res = ts_ens.correlation(ts_target)
        signif_list = corr_res.signif
        for signif in signif_list:
            assert signif is True

        assert np.size(corr_res.p) == np.size(ts_list1)


    def test_correlation_t3(self):
        '''Test for EnsembleSeries.correlation() when the target is an EnsembleSeries with more Series
        '''
        nt = 100
        t0, v0 = gen_colored_noise(nt=nt)
        t0, noise = gen_normal(nt=nt)

        ts0 = pyleo.Series(time=t0, value=v0)
        ts1 = pyleo.Series(time=t0, value=v0+noise)
        ts2 = pyleo.Series(time=t0, value=v0+2*noise)
        ts3 = pyleo.Series(time=t0, value=v0+1/2*noise)
        ts4 = pyleo.Series(time=t0, value=v0+3/2*noise)

        ts_list1 = [ts0, ts1]
        ts_list2 = [ts2, ts3, ts4]

        ts_ens = pyleo.EnsembleSeries(ts_list1)
        ts_target = pyleo.EnsembleSeries(ts_list2)

        corr_res = ts_ens.correlation(ts_target)
        signif_list = corr_res.signif
        for signif in signif_list:
            assert signif is True

        assert np.size(corr_res.p) == np.size(ts_list1)

    def test_plot_envelope_t0(self):
        ''' Test EnsembleSeries.plot_envelope() on a list of colored noise
        '''
        nn = 30 # number of noise realizations
        nt = 500
        series_list = []

        signal = gen_ts(model='colored_noise', nt=nt, alpha=1.0).standardize()

        noise = np.random.randn(nt,nn)

        for idx in range(nn):  # noise
            ts = pyleo.Series(time=signal.time, value=signal.value+noise[:,idx])
            series_list.append(ts)

        ts_ens = pyleo.EnsembleSeries(series_list)

        fig, ax = ts_ens.plot_envelope(curve_lw=1.5)
        pyleo.closefig(fig)

    def test_plot_traces_t0(self):
        ''' Test EnsembleSeries.plot_traces() on a list of colored noise
        '''
        nn = 30 # number of noise realizations
        nt = 500
        series_list = []

        signal = gen_ts(model='colored_noise', nt=nt, alpha=1.0).standardize()
        noise = np.random.randn(nt, nn)

        for idx in range(nn):  # noise
            ts = pyleo.Series(time=signal.time, value=signal.value+noise[:,idx])
            series_list.append(ts)

        ts_ens = pyleo.EnsembleSeries(series_list)

        fig, ax = ts_ens.plot_traces(alpha=0.2, num_traces=8)
        pyleo.closefig(fig)

class TestUIEnsembleSeriesHistplot():
    def test_histplot_t0(self):
        '''Test for EnsembleSeries.histplot()
        '''
        nn = 30 # number of noise realizations
        nt = 500
        series_list = []

        signal = gen_ts(model='colored_noise', nt=nt, alpha=1.0).standardize()
        noise = np.random.randn(nt, nn)

        for idx in range(nn):  # noise
            ts = pyleo.Series(time=signal.time, value=signal.value+noise[:,idx])
            series_list.append(ts)

        ts_ens = pyleo.EnsembleSeries(series_list)

        ts_ens.histplot()

class TestUIEnsembleSeriesDistplot():
    def test_histplot_t0(self):
        '''Test for EnsembleSeries.distplot()
        '''
        nn = 30 # number of noise realizations
        nt = 500
        series_list = []

        signal = gen_ts(model='colored_noise', nt=nt, alpha=1.0).standardize()
        noise = np.random.randn(nt, nn)

        for idx in range(nn):  # noise
            ts = pyleo.Series(time=signal.time, value=signal.value+noise[:,idx])
            series_list.append(ts)

        ts_ens = pyleo.EnsembleSeries(series_list)

        ts_ens.histplot()
